import sys
from collections import Counter, OrderedDict

import numpy as np
from scipy.spatial.distance import cosine

from spine.ana.base import AnaBase
from spine.math.metrics import adjusted_rand_score

__all__ = []


# @write_to(['particle_pairs_t2r', 'particle_pairs_r2t'])
def select_particle_pairs(data_blob, res, **kwargs):
    """
    Select particle pairs for logging (only for mpv/nu interactions)
    """

    particles_t2r, particles_r2t = [], []

    particle_fieldnames = kwargs["logger"].get("particles", {})
    pixel_threshold = kwargs.get("pixel_threshold", -1)
    semantic_label = kwargs.get("semantic_label", 1)
    only_primaries = kwargs.get("only_primaries", True)

    image_idxs = data_blob["index"]
    meta = data_blob["meta"][0]

    for idx, index in enumerate(image_idxs):

        index_dict = {
            "Iteration": kwargs["iteration"],
            "Index": index,
            # 'file_index': data_blob['file_index'][idx]
        }

        # 1. Match Particles
        # pmatches, pcounts = res['matched_particles_t2r'][idx], res['particle_match_overlap_t2r'][idx]
        pmatches = []
        for p in res["matched_particles_t2r"][idx]:
            if p[0].size < pixel_threshold:
                continue
            if p[0].semantic_type != semantic_label:
                continue
            if only_primaries and not p[0].is_primary:
                continue
            if p[0].nu_id == 1:
                pmatches.append(p)

        if len(pmatches) > 0:
            # pindices = [i for i, p in enumerate(res['matched_particles_t2r'][idx]) if p[0].nu_id == 1]
            # pcounts = [pm for i, pm in enumerate(pmatches) if i == pindices[i]]

            # 3. Process particle level information
            particle_logger = ParticleLogger(particle_fieldnames, meta=meta)
            particle_logger.prepare()

            for i, p1 in enumerate(pmatches):
                for j, p2 in enumerate(pmatches):
                    if i < j:
                        true_1, true_2 = p1[0], p2[0]
                        reco_1, reco_2 = p1[1], p2[1]

                        reco_pair = [reco_1, reco_2]
                        true_pair = [true_1, true_2]

                        # Colinearity
                        if (
                            np.isinf(true_1.truth_start_dir).any()
                            or np.isinf(true_2.truth_start_dir).any()
                        ):
                            dist = np.inf
                        else:
                            dist = cosine(
                                true_1.truth_start_dir, true_2.truth_start_dir
                            )
                        # Compute overlap
                        mean_iou, max_iou = twopoint_iou(reco_pair, true_pair)

                        index_dict["cosine_dist"] = dist
                        index_dict["twopoint_mean_iou"] = mean_iou
                        index_dict["twopoint_max_iou"] = max_iou

                        true_p1_dict = particle_logger.produce(
                            true_1, mode="true", prefix="p1"
                        )
                        true_p2_dict = particle_logger.produce(
                            true_2, mode="true", prefix="p2"
                        )

                        pred_p1_dict = particle_logger.produce(
                            reco_1, mode="reco", prefix="p1"
                        )
                        pred_p2_dict = particle_logger.produce(
                            reco_2, mode="reco", prefix="p2"
                        )

                        part_dict = OrderedDict()
                        part_dict.update(index_dict)
                        part_dict.update(true_p1_dict)
                        part_dict.update(true_p2_dict)
                        part_dict.update(pred_p1_dict)
                        part_dict.update(pred_p2_dict)
                        particles_t2r.append(part_dict)

        # 1. Match Interactions and log interaction-level information
        pmatches = []
        for p in res["matched_particles_r2t"][idx]:
            if p[1] is None:
                continue
            if p[1].size < pixel_threshold:
                continue
            if p[1].semantic_type != semantic_label:
                continue
            if only_primaries and not p[1].is_primary:
                continue
            if p[1].nu_id == 1:
                pmatches.append(p)
        # pmatches = [p for p in res['matched_particles_r2t'][idx] if (p[1] is not None and p[1].nu_id == 1)]

        if len(pmatches) > 0:
            # pindices = [i for i, p in enumerate(res['matched_particles_r2t'][idx]) if (p[1] is not None and p[1].nu_id == 1)]
            # pcounts = [pm for i, pm in enumerate(pmatches) if i == pindices[i]]

            # 3. Process particle level information
            particle_logger = ParticleLogger(particle_fieldnames, meta=meta)
            particle_logger.prepare()

            for i, p1 in enumerate(pmatches):
                for j, p2 in enumerate(pmatches):
                    if i < j:
                        true_1, true_2 = p1[1], p2[1]
                        reco_1, reco_2 = p1[0], p2[0]

                        reco_pair = [reco_1, reco_2]
                        true_pair = [true_1, true_2]

                        # Colinearity
                        if (
                            np.isinf(true_1.truth_start_dir).any()
                            or np.isinf(true_2.truth_start_dir).any()
                        ):
                            dist = np.inf
                        else:
                            dist = cosine(
                                true_1.truth_start_dir, true_2.truth_start_dir
                            )
                        # Compute overlap
                        score = twopoint_iou(reco_pair, true_pair)

                        index_dict["cosine_dist"] = dist
                        index_dict["twopoint_iou"] = score

                        true_p1_dict = particle_logger.produce(
                            true_1, mode="true", prefix="p1"
                        )
                        true_p2_dict = particle_logger.produce(
                            true_2, mode="true", prefix="p2"
                        )

                        pred_p1_dict = particle_logger.produce(
                            reco_1, mode="reco", prefix="p1"
                        )
                        pred_p2_dict = particle_logger.produce(
                            reco_2, mode="reco", prefix="p2"
                        )

                        part_dict = OrderedDict()
                        part_dict.update(index_dict)
                        part_dict.update(true_p1_dict)
                        part_dict.update(true_p2_dict)
                        part_dict.update(pred_p1_dict)
                        part_dict.update(pred_p2_dict)
                        particles_r2t.append(part_dict)

    return [particles_t2r, particles_r2t]



#!/usr/bin/env python3
"""Mark all bad LArCV ROOT files before merging them with hadd."""

import argparse

import numpy as np
from larcv import larcv  # pylint: disable=W0611
from ROOT import TFile  # pylint: disable=E0611
from tqdm import tqdm


def main(source, source_list, output):
    """Checks the validity of a LArCV root file.

    This script loops over all TTrees in a given ROOT file and check that they
    have the same, non-zero, number of entries.

    Produces a list of bad files in 'bad_files.txt' (one per line) that can then
    be used to move/remove these bad files before doing hadd. For example using:

    .. code-block:: bash

        $ for file in $(cat bad_files.txt); do mv "$file" bad_files/; done

    Parameters
    ----------
    source : List[str]
        List of paths to the input files
    source_list : str
        Path to a text file containing a list of data file paths
    output : str
        Path to the output text file with the list of bad files
    """
    # If using source list, read it in
    if source_list is not None:
        with open(source_list, "r", encoding="utf-8") as f:
            source = f.read().splitlines()

    # Initialize the output text file
    out_file = open(output, "w", encoding="utf-8")

    # Loop over the list of files in the input, count the tree entries for each
    print(f"\nCounting entries in every tree of {len(source)} files.")
    keys_list, unique_counts = [], []
    for file_path in tqdm(source):
        # Count the number of entries in each tree
        try:
            f = TFile(file_path)
        except OSError:
            keys_list.append([])
            unique_counts.append([])
            continue

        keys = [key.GetName() for key in f.GetListOfKeys()]
        trees = [f.Get(key) for key in keys]
        num_entries = [tree.GetEntries() for tree in trees]
        f.Close()

        keys_list.append(keys)
        unique_counts.append(np.unique(num_entries))

    # Get the all the unique tree names encountered in the list of files
    all_keys = np.unique(np.concatenate(keys_list))

    # Loop over the list of keys/counts for each file in the input
    print(f"\nChecking validity of {len(source)} file(s).")
    bad_files = []
    for idx, file_path in enumerate(tqdm(source)):
        # Check that there is only one entry count and it's non-zero, and
        # that the list of keys matches expectation
        if (
            len(unique_counts[idx]) != 1
            or unique_counts[idx][0] < 1
            or (set(keys_list[idx]) != set(all_keys))
        ):
            tqdm.write(f"- Bad file: {file_path}")
            out_file.write(f"{file_path}\n")
            bad_files.append(file_path)

    print(f"\nFound {len(bad_files)} bad files.")

    # Close text file
    out_file.close()


if __name__ == "__main__":
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description="Check dataset validity")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--source",
        "-s",
        help="Path or list of paths to data files",
        type=str,
        nargs="+",
    )
    group.add_argument(
        "--source-list", "-S", help="Path to a text file of data file paths", type=str
    )

    parser.add_argument(
        "--output",
        "-o",
        help="Path to the output text file with the bad list",
        type=str,
        required=True,
    )

    args = parser.parse_args()

    # Execute the main function
    main(args.source, args.source_list, args.output)



"""Contains a reader class dedicated to loading data from HDF5 files."""

from dataclasses import fields

import h5py
import numpy as np
import yaml

import spine.data
from spine.utils.docstring import inherit_docstring
from spine.utils.logger import logger

from .base import ReaderBase

__all__ = ["HDF5Reader"]


@inherit_docstring(ReaderBase)
class HDF5Reader(ReaderBase):
    """Class which reads information stored in HDF5 files.

    This class inherits from the :class:`ReaderBase` class. It provides
    methods to load HDF5 files and extract their data products. The files
    must be structured as follows:
      - An `events` dataset with all the region references
      - One dataset per data product corresponding to each region reference in
        the `events` dataset
    """

    name = "hdf5"

    def __init__(
        self,
        file_keys,
        limit_num_files=None,
        max_print_files=10,
        n_entry=None,
        n_skip=None,
        entry_list=None,
        skip_entry_list=None,
        run_event_list=None,
        skip_run_event_list=None,
        create_run_map=False,
        build_classes=True,
        skip_unknown_attrs=False,
        run_info_key="run_info",
        allow_missing=False,
    ):
        """Initalize the HDF5 file reader.

        Parameters
        ----------
        file_keys : list
            List of paths to the HDF5 files to be read
        limit_num_files : int, optional
            Integer limiting number of files to be taken per data directory
        max_print_files : int, default 10
            Maximum number of loaded file names to be printed
        n_entry : int, optional
            Maximum number of entries to load
        n_skip : int, optional
            Number of entries to skip at the beginning
        entry_list : list
            List of integer entry IDs to add to the index
        skip_entry_list : list
            List of integer entry IDs to skip from the index
        run_event_list: list((int, int, int)), optional
            List of (run, subrun, event) triplets to add to the index
        skip_run_event_list: list((int, int, int)), optional
            List of (run, subrun, event) triplets to skip from the index
        create_run_map : bool, default False
            Initialize a map between (run, subrun, event) triplets and entries.
            For large files, this can be quite expensive (must load every entry).
        build_classes : bool, default True
            If the stored object is a class, build it back
        skip_unknown_attrs : bool, default False
            If `True`, allow a loaded object to have unrecognized attributes.
            This allows backward compatibility with old files, but use with
            extreme caution, as this might hide a fundamental issue with your code.
        run_info_key : str, default 'run_info'
            Name of the data product which contains the run info of the event
        allow_missing : bool, default False
            If `True`, allows missing entries in the entry or event list
        """
        # Process the list of files
        self.process_file_paths(file_keys, limit_num_files, max_print_files)

        # If an entry list is requested based on run/subrun/event ID, create map
        if run_event_list is not None or skip_run_event_list is not None:
            create_run_map = True

        # Loop over the input files, build a map from index to file ID
        self.num_entries = 0
        self.file_index = []
        self.file_offsets = np.empty(len(self.file_paths), dtype=np.int64)
        self.run_info = None if not create_run_map else []
        for i, path in enumerate(self.file_paths):
            with h5py.File(path, "r") as in_file:
                # Check that there are events in the file
                assert "events" in in_file, "File does not contain an event tree"

                # If requested, register the (run, subrun, event) information
                if create_run_map:
                    assert (
                        run_info_key in in_file
                    ), f"Must provide {run_info_key} to create run map"
                    info = in_file[run_info_key]
                    for r, s, e in zip(info["run"], info["subrun"], info["event"]):
                        self.run_info.append((r, s, e))

                # Update the total number of entries
                num_entries = len(in_file["events"])
                self.file_index.append(i * np.ones(num_entries, dtype=np.int64))
                self.file_offsets[i] = self.num_entries
                self.num_entries += num_entries

        # Dump the number of entries to load
        logger.info("Total number of entries in the file(s): %d\n", self.num_entries)

        # Concatenate the file indexes into one
        self.file_index = np.concatenate(self.file_index)

        # Turn file index and run info into np.ndarray
        if self.run_info is not None:
            self.run_info = np.vstack(self.run_info)

        # Process the run information
        self.process_run_info()

        # Process the entry list
        self.process_entry_list(
            n_entry,
            n_skip,
            entry_list,
            skip_entry_list,
            run_event_list,
            skip_run_event_list,
            allow_missing,
        )

        # Store other attributes
        self.build_classes = build_classes
        self.skip_unknown_attrs = skip_unknown_attrs

        # Process the configuration used to produce the HDF5 file
        self.cfg = self.process_cfg()

        # Process the SPINE version used to produced the HDF5 file
        self.version = self.process_version()

    def process_cfg(self):
        """Fetches the SPINE configuration used to produce the HDF5 file.

        Returns
        -------
        dict
            Configuration dictionary
        """
        # Fetch the string-form configuration
        with h5py.File(self.file_paths[0], "r") as in_file:
            cfg_str = in_file["info"].attrs["cfg"]

        # Attempt to parse it (need try for now for SPINE versions < v0.4.0)
        try:
            cfg = yaml.safe_load(cfg_str)
        except:
            return None

        return cfg

    def process_version(self):
        """Returns the SPINE release version used to produce the HDF5 file.

        Returns
        -------
        str
            SPINE release tag
        """
        # Fetch the string-form configuration
        with h5py.File(self.file_paths[0], "r") as in_file:
            version = in_file["info"].attrs["version"]

        return version

    def get(self, idx):
        """Returns a specific entry in the file.

        Parameters
        ----------
        idx : int
            Integer entry ID to access

        Returns
        -------
        data : dict
            Ditionary of data products corresponding to one event
        """
        # Get the appropriate entry index
        assert idx < len(self.entry_index)
        file_idx = self.get_file_index(idx)
        entry_idx = self.get_file_entry_index(idx)

        # Use the event tree to find out what needs to be loaded
        data = {"file_index": file_idx, "file_entry_index": entry_idx}
        with h5py.File(self.file_paths[file_idx], "r") as in_file:
            event = in_file["events"][entry_idx]
            for key in event.dtype.names:
                self.load_key(in_file, event, data, key)

        # Use the global index, not the one read from file
        data["index"] = np.int64(idx)

        return data

    def load_key(self, in_file, event, data, key):
        """Fetch a specific key for a specific event.

        Parameters
        ----------
        in_file : h5py.File
            HDF5 file instance
        event : dict
            Dictionary of objects that make up one event
        data : dict
            Dictionary of data products corresponding to one event
        key: str
            Name of the dataset in the entry
        """
        # The event-level information is a region reference: fetch it
        region_ref = event[key]
        if isinstance(in_file[key], h5py.Dataset):
            if not in_file[key].dtype.names:
                # If the reference points at a simple dataset, return
                data[key] = in_file[key][region_ref]
                if in_file[key].attrs["scalar"]:
                    data[key] = data[key][0]
                if len(in_file[key].shape) > 1:
                    data[key] = data[key].reshape(-1, in_file[key].shape[1])

            else:
                # If the dataset has multiple attributes, it contains an object.
                # Start by fetching the appropriate class to rebuild
                array = in_file[key][region_ref]
                class_name = in_file[key].attrs["class_name"]
                obj_class = getattr(spine.data, class_name)

                # If needed, get the list of recognized attributes
                if self.skip_unknown_attrs:
                    known_attrs = [f.name for f in fields(obj_class)]

                # Load the object
                names = array.dtype.names
                data[key] = []
                for i, el in enumerate(array):
                    # Fetch the list of key/value pairs, filter if requested
                    if self.skip_unknown_attrs:
                        obj_dict = {}
                        for i, k in enumerate(names):
                            if k in known_attrs:
                                obj_dict[k] = el[i]
                    else:
                        obj_dict = dict(zip(names, el))

                    # Rebuild an instance of the object class, if requested
                    if self.build_classes:
                        data[key].append(obj_class(**obj_dict))
                    else:
                        data[key].append(obj_dict)

                if in_file[key].attrs["scalar"]:
                    data[key] = data[key][0]

        else:
            # If the reference points at a group, unpack
            el_refs = in_file[key]["index"][region_ref].flatten()
            if len(in_file[key]["index"].shape) == 1:
                ret = np.empty(len(el_refs), dtype=object)
                ret[:] = [in_file[key]["elements"][r] for r in el_refs]
                if len(in_file[key]["elements"].shape) > 1:
                    for i in range(len(el_refs)):
                        ret[i] = ret[i].reshape(-1, in_file[key]["elements"].shape[1])

            else:
                ret = [in_file[key][f"element_{i}"][r] for i, r in enumerate(el_refs)]
                for i in range(len(el_refs)):
                    if len(in_file[key][f"element_{i}"].shape) > 1:
                        ret[i] = ret[i].reshape(
                            -1, in_file[key][f"element_{i}"].shape[1]
                        )

            data[key] = ret



"""Match objects and their label counterparts and vice versa."""

from dataclasses import dataclass

import numba as nb
import numpy as np

import spine.utils.match
from spine.post.base import PostBase

__all__ = ["MatchProcessor"]


class MatchProcessor(PostBase):
    """Does the matching between reconstructed and true objects."""

    # Name of the post-processor (as specified in the configuration)
    name = "match"

    def __init__(
        self,
        fragment=None,
        particle=None,
        interaction=None,
        truth_point_mode="points",
        **kwargs,
    ):
        """Initializes the matching post-processor.

        Parameters
        ----------
        fragment: Union[bool, dict], optional
            Matching flag or configuration for fragments
        particle: Union[bool, dict], optional
            Matching flag or configuration for particles
        interaction: Union[bool, dict], optional
            Matching flag or configuration for interactions
        truth_point_mode : str, default 'points'
            Type of truth points to use to do the matching
        **kwargs : dict, optional
            Matching parameters shared between all matching processes
        """
        # Initialize the necessary matchers
        configs = {
            "fragment": fragment,
            "particle": particle,
            "interaction": interaction,
        }
        keys = {}
        self.matchers = {}
        for key, cfg in configs.items():
            if cfg is not None and cfg != False:
                # Initialize the matcher
                if isinstance(cfg, bool):
                    cfg = {}
                self.matchers[key] = self.Matcher(**cfg, **kwargs)

                # If any matcher includes ghost points, must load meta
                if self.matchers[key].ghost:
                    keys["meta"] = True

        assert len(
            self.matchers
        ), "Must specify one of 'fragment', 'particle' or 'interaction'."

        # Update the set of keys necessary for this post-processor
        self.update_keys(keys)

        # Initialize the parent class
        super().__init__(list(self.matchers.keys()), "both", truth_point_mode)

    @dataclass
    class Matcher:
        """Simple data class to store matching methods per object.

        Attributes
        ----------
        fn : function
            Function which computes overlaps between pairs of objects
        match_mode : str, defualt 'both'
            Matching mode. One of 'reco_to_truth', 'truth_to_reco' or 'both'
        overlap_mode : str, default 'iou'
            Overlap estimatation method. One of 'count', 'iou', 'dice', 'chamfer'
        min_overlap : float, default 0.
            Overlap value above which a pair is considered a match
        weight_overlap : bool, default False
            Whether to weight the overlap metric
        ghost : bool, default False
            Whether a deghosting process was applied (in which case the indexes
            of the reco and the truth particles do not align)
        """

        fn: object = None
        match_mode: str = "both"
        overlap_mode: str = "iou"
        min_overlap: float = 0.0
        weight_overlap: bool = False
        ghost: bool = False

        # Valid match modes
        _match_modes = ["reco_to_truth", "truth_to_reco", "both", "all"]

        # Valid overlap modes
        _overlap_modes = ["count", "iou", "dice", "chamfer"]

        def __post_init__(self):
            """Check that the values provided are valid."""
            # Check match mode
            assert self.match_mode in self._match_modes, (
                f"Invalid matching mode: {self.match_mode}. Must be one "
                f"of {self._match_modes}."
            )

            # Check the overlap mode
            assert self.overlap_mode in self._overlap_modes, (
                f"Invalid overlap computation mode: {self.overlap_mode}. "
                f"Must be one of {self._overlap_modes}."
            )

            # Check that the overlap mode and weighting are compatible
            assert not self.weight_overlap or overlap in [
                "iou",
                "dice",
            ], "Only IoU and Dice-based overlap functions can be weighted."

            # Initialize the match overlap function
            prefix = "overlap" if not self.weight_overlap else "overlap_weighted"
            self.fn = getattr(spine.utils.match, f"{prefix}_{self.overlap_mode}")

    def process(self, data):
        """Match all the requested objects in one entry.

        Parameters
        ----------
        data: dict
            Dictionary of data products
        """
        # Loop over the matchers
        result = {}
        for name, matcher in self.matchers.items():
            # Fetch the required data products
            reco_objs = data[f"reco_{name}s"]
            truth_objs = data[f"truth_{name}s"]

            # Fetch the metadata, if needed
            meta = None
            if matcher.ghost:
                meta = data["meta"]

            # Pass it to the individual processor
            res_one = self.process_single(reco_objs, truth_objs, matcher, name, meta)
            result.update(**res_one)

        return result

    def process_single(self, reco_objs, truth_objs, matcher, name, meta=None):
        """Match all the requested objects in a single category.

        Parameters
        ----------
        reco_objs : List[object]
            List of reconstructed objects
        truth_objs : List[object]
            List of truth objects
        matcher : MatchProcessor.Matcher
            Matching method and function
        name : str
            Object type name
        meta : Meta, optional
            Metadata information to convert position to index
        """
        # Convert the object list into an index/coordinate list
        if matcher.overlap_mode != "chamfer":
            # For overlap matches, use pixel indexes (faster)
            if not matcher.ghost or self.truth_point_mode == "points_adapt":
                # The indexes of reco and truth point to the same point set
                reco_input = nb.typed.List([self.get_index(p) for p in reco_objs])
                truth_input = nb.typed.List([self.get_index(p) for p in truth_objs])

            else:
                # The indexes of reco and truth point to different point sets.
                # In this case, convert the positions to indexes
                reco_input = []
                for p in reco_objs:
                    coords = self.get_points(p)
                    if p.units != "px":
                        coords = meta.to_px(coords, floor=True)
                    reco_input.append(meta.index(coords))

                truth_input = []
                for p in truth_objs:
                    coords = self.get_points(p)
                    if p.units != "px":
                        coords = meta.to_px(coords, floor=True)
                    truth_input.append(meta.index(coords))

        else:
            # For the chamfer distance, simply use the point positions
            reco_input = nb.typed.List([self.get_points(p) for p in reco_objs])
            truth_input = nb.typed.List([self.get_points(p) for p in truth_objs])

        # Pass lists to the matching function to compute overlaps
        if len(reco_input) and len(truth_input):
            ovl_matrix = matcher.fn(reco_input, truth_input)
        else:
            ovl_matrix = np.empty((len(reco_input), len(truth_input)))

        # Make the overlap selection cut, if requested
        if matcher.overlap_mode != "chamfer":
            ovl_valid = ovl_matrix > matcher.min_overlap
        else:
            ovl_valid = ovl_matrix < match.min_overlap

        # Produce matches
        result = {}
        if matcher.match_mode != "truth_to_reco":
            pairs, overlaps = self.generate_matches(
                reco_objs, truth_objs, ovl_matrix, ovl_valid
            )
            result[f"{name}_matches_r2t"] = pairs
            result[f"{name}_matches_r2t_overlap"] = overlaps

        if matcher.match_mode != "reco_to_truth":
            pairs, overlaps = self.generate_matches(
                truth_objs, reco_objs, ovl_matrix.T, ovl_valid.T
            )
            result[f"{name}_matches_t2r"] = pairs
            result[f"{name}_matches_t2r_overlap"] = overlaps

        return result

    @staticmethod
    def generate_matches(source_objs, target_objs, ovl_matrix, ovl_valid):
        """Generate pairs for a srt of sources and targets.

        Parameters
        ----------
        source_objs : List[object]
            (N) List of source objects
        target_objs : List[object]
            (M) List of truth objects
        ovl_matrix : np.ndarray
            (N, M) Matrix of overlap values
        ovl_valid : np.ndarray
            (N, M) Matrix of overlap validity

        Returns
        -------
        pairs : List[tuple]
            (N) List of (source, target) matched pairs (best match only)
        overlaps : List[float]
            (N) List of overlap between each source and the best matched target
        """
        # Build the matches based on the threshold
        pairs, pair_overlaps = [], []
        for i, s in enumerate(source_objs):
            # Get the list of valid matches
            match_idxs = np.where(ovl_valid[i])[0]
            if not len(match_idxs):
                # If there are no matches, fill dummy values
                s.is_matched = False
                s.match_ids = np.empty(0, dtype=np.int64)
                s.match_overlaps = np.empty(0, dtype=np.float32)

                pairs.append((s, None))
                pair_overlaps.append(-1.0)

            else:
                # If there are matches, order them by decreasing overlap
                overlaps = ovl_matrix[i, match_idxs]
                perm = np.argsort(overlaps)[::-1]
                s.is_matched = True
                s.match_ids = match_idxs[perm]
                s.match_overlaps = overlaps[perm]

                best_idx = s.match_ids[0]
                pairs.append((s, target_objs[best_idx]))
                pair_overlaps.append(s.match_overlaps[0])

        # Fill the match lists
        return pairs, pair_overlaps



"""Module with a general-purpose geometry class.

This class supports the storage of:
- TPC boundaries
- Optical detector shape/locations
- CRT detector shape/locations

It also provides a plethora of useful functions to query the geometry.
"""

import os
import pathlib
from dataclasses import dataclass

import numpy as np
import yaml
from scipy.spatial.distance import cdist

from .detector import *


@dataclass
class Geometry:
    """Handles all geometry functions for a collection of box-shaped TPCs with
    a arbitrary set of optical detectors organized in optical volumes and CRT
    planes.

    Attributes
    ----------
    tpc : TPCDetector
        TPC detector properties
    optical : OptDetector, optional
        Optical detector properties
    crt : CRTDetector, optional
        CRT detector properties
    """

    tpc: TPCDetector
    optical: OptDetector = None
    crt: CRTDetector = None

    def __init__(self, detector=None, file_path=None):
        """Initializes a detector geometry object.

        The geometry configuration file is a YAML file which contains all the
        necessary information to construct the physical boundaries of the
        a detector (TPC size, positions, etc.) and (optionally) the set
        of optical detectors and CRTs.

        If the detector is already supported, the name is sufficient.
        Supported: 'icarus', 'sbnd', '2x2', '2x2_single', 'ndlar'

        Parameters
        ----------
        detector : str, optional
            Name of a recognized detector to the geometry from
        file_path : str, optional
            Path to a `.yaml` geometry configuration
        """
        # Check that we are provided with either a detector name or a file
        assert (detector is not None) ^ (file_path is not None), (
            "Must provide either a `detector` name or a geometry "
            "`file_path`, not neither, not both."
        )

        # If a detector name is provided, find the corresponding geometry file
        if detector is not None:
            path = pathlib.Path(__file__).parent
            file_path = os.path.join(
                path, "source", f"{detector.lower()}_geometry.yaml"
            )

        # Check that the geometry configuration file exists
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"Could not find the geometry file: {file_path}")

        # Load the geometry file, parse it
        with open(file_path, "r", encoding="utf-8") as cfg_yaml:
            cfg = yaml.safe_load(cfg_yaml)

        self.parse_configuration(**cfg)

        # Initialize place-holders for the containment conditions to be defined
        # by the end-user using :func:`define_containment_volumes`
        self._cont_volumes = None
        self._cont_limits = None
        self._cont_use_source = False

    def parse_configuration(self, tpc, optical=None, crt=None):
        """Parse the geometry configuration.

        Parameters
        ----------
        tpc : dict
            Detector boundary configuration
        optical : dict, optional
            Optical detector configuration
        crt : dict, optional
            CRT detector configuration
        """
        # Load the charge detector boundaries
        self.tpc = TPCDetector(**tpc)

        # Load the optical detectors
        if optical is not None:
            self.parse_optical(**optical)

        # Load the CRT detectors
        if crt is not None:
            self.crt = CRTDetector(**crt)

    def parse_optical(self, volume, **optical):
        """Parse the optical detector configuration.

        Parameters
        ----------
        volume : str
            Optical volume boundaries (one of 'tpc' or 'module')
        **optical : dict
            Reset of the optical detector configuration
        """
        # Get the number of optical volumes based on the the volume type
        assert volume in [
            "module",
            "tpc",
        ], "Optical detector positions must be provided by TPC or module."

        if volume == "module":
            offsets = [module.center for module in self.tpc.modules]
        else:
            offsets = [chamber.center for chamber in self.tpc.chambers]

        # Initialize the optical detector object
        self.optical = OptDetector(volume, offsets, **optical)

    def get_sources(self, sources):
        """Converts logical TPC indexes to physical TPC indexes.

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of logical [module ID, tpc ID] pairs, one per point

        Returns
        ----------
        np.ndarray
            (N, 2) Array of physical [module ID, tpc ID] pairs, one per point
        """
        # If logical and physical TPCs are aligned, nothing to do
        if self.tpc.det_ids is None:
            return sources.astype(int)

        # Otherwise, map logical to physical
        sources = np.copy(sources)
        sources[:, 1] = self.tpc.det_ids[sources[:, 1].astype(int)]

        return sources

    def get_chambers(self, sources):
        """Converts logical TPC indexes to unique chamber indexes.

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of logical [module ID, tpc ID] pairs, one per point

        Returns
        ----------
        np.ndarray
            (N) Array of physical chamber indexes, one per point
        """
        # Convert the physical indexes to unique chamber indexes
        sources = self.get_sources(sources)
        return sources[:, 0] * self.tpc.num_chambers_per_module + sources[:, 1]

    def get_contributors(self, sources):
        """Gets the list of [module ID, tpc ID] pairs that contributed to an
        object, as defined in this geometry.

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of [module ID, tpc ID] pairs, one per point

        Returns
        -------
        List[np.ndarray]
            (2, N_t) Pair of arrays: the first contains the list of
            contributing modules, the second of contributing tpcs.
        """
        # Fetch the list of unique logical [module ID, tpc ID] pairs
        sources = np.unique(sources, axis=0)

        # If the logical TPCs differ from the physical TPCs, convert
        if self.tpc.det_ids is not None:
            sources = self.get_sources(sources)
            sources = np.unique(sources, axis=0)

        # Return as a list of physical [module ID, tpc ID] pairs
        return list(sources.T)

    def get_volume_index(self, sources, module_id, tpc_id=None):
        """Gets the list of indices of points that belong to a certain
        detector volume (module or individual TPC).

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of [module ID, tpc ID] pairs, one per point
        module_id : int
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module. If not specified, the volume
            offsets are estimated w.r.t. the module itself

        Returns
        -------
        np.ndarray
            (N) Index of points that belong to the requested detector volume
        """
        # If the logical TPCs differ from the physical TPCs, convert
        sources = self.get_sources(sources)

        # Compute and return the index
        if tpc_id is None:
            return np.where(sources[:, 0] == module_id)[0]
        else:
            return np.where((sources == [module_id, tpc_id]).all(axis=-1))[0]

    def get_closest_tpc(self, points):
        """For each point, find the ID of the closest TPC.

        There is a natural assumption that all TPCs are boxes of identical
        sizes, so that the relative proximitity of a point to a TPC is
        equivalent to its proximity to the TPC center.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        np.ndarray
            (N) List of TPC indexes, one per input point
        """
        # Get the TPC centers
        centers = np.asarray([chamber.center for chamber in self.tpc.chambers])

        # Compute the pair-wise distances between points and TPC centers
        dists = cdist(points, centers)

        # Return the closest center index as the closest centers
        return np.argmin(dists, axis=1)

    def get_closest_module(self, points):
        """For each point, find the ID of the closest module.

        There is a natural assumption that all modules are boxes of identical
        sizes, so that the relative proximitity of a point to a module is
        equivalent to its proximity to the module center.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        np.ndarray
            (N) List of module indexes, one per input point
        """
        # Get the module centers
        centers = np.asarray([module.center for module in self.tpc.modules])

        # Compute the pair-wise distances between points and module centers
        dists = cdist(points, centers)

        # Return the closest center index as the closest centers
        return np.argmin(dists, axis=1)

    def get_closest_tpc_indexes(self, points):
        """For each TPC, get the list of points that live closer to it than any
        other TPC in the detector.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        List[np.ndarray]
            List of index of points that belong to each TPC
        """
        # Start by finding the closest TPC to each of the points
        closest_ids = self.get_closest_tpc(points)

        # For each TPC, append the list of point indices associated with it
        tpc_indexes = []
        for t in range(self.tpc.num_chambers):
            tpc_indexes.append(np.where(closest_ids == t)[0])

        return tpc_indexes

    def get_closest_module_indexes(self, points):
        """For each module, get the list of points that live closer to it
        than any other module in the detector.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        List[np.ndarray]
            List of index of points that belong to each module
        """
        # Start by finding the closest TPC to each of the points
        closest_ids = self.get_closest_module(points)

        # For each module, append the list of point indices associated with it
        module_indexes = []
        for m in range(self.tpc.num_modules):
            module_indexes.append(np.where(closest_ids == m)[0])

        return module_indexes

    def get_volume_offsets(self, points, module_id, tpc_id=None):
        """Compute how far each point is from a certain detector volume.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) : Point coordinates
        module_id : int
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module. If not specified, the volume
            offsets are estimated w.r.t. the module itself

        Returns
        -------
        np.ndarray
            (N, 3) Offsets w.r.t. to the volume boundaries
        """
        # Compute the axis-wise distances of each point to each boundary
        idx = module_id if tpc_id is None else (module_id, tpc_id)
        boundaries = self.tpc[idx].boundaries
        dists = points[..., None] - boundaries

        # If a point is between two boundaries, the distance is 0. If it is
        # outside, the distance is that of the closest boundary
        signs = (np.sign(dists[..., 0]) + np.sign(dists[..., 1])) / 2
        offsets = signs * np.min(np.abs(dists), axis=-1)

        return offsets

    def get_min_volume_offset(self, points, module_id, tpc_id=None):
        """Get the minimum offset to apply to a point cloud to bring it
        within the boundaries of a volume.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) : Point coordinates
        module_id : int
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module. If not specified, the volume
            offsets are estimated w.r.t. the module itself

        Returns
        -------
        np.ndarray
            (3) Offsets w.r.t. to the volume location
        """
        # Compute the distance for each point, get the maximum necessary offset
        offsets = self.get_volume_offsets(points, module_id, tpc_id)
        offsets = offsets[np.argmax(np.abs(offsets), axis=0), np.arange(3)]

        return offsets

    def translate(self, points, source_id, target_id, factor=None):
        """Moves a point cloud from one module to another one

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates
        source_id: int
            Module ID from which to move the point cloud
        target_id : int
            Module ID to which to move the point cloud
        factor : Union[float, np.ndarray], optional
            Multiplicative factor to apply to the offset. This is necessary if
            the points are not expressed in detector coordinates

        Returns
        -------
        np.ndarray
            (N, 3) Set of translated point coordinates
        """
        # If the source and target are the same, nothing to do here
        if target_id == source_id:
            return np.copy(points)

        # Fetch the inter-module shift
        offset = self.tpc[target_id].center - self.tpc[source_id].center
        if factor is not None:
            offset *= factor

        # Translate
        return points + offset

    def split(self, points, target_id, sources=None, meta=None):
        """Migrate all points to a target module, organize them by module ID.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates
        target_id : int
            Module ID to which to move the point cloud
        sources : np.ndarray, optional
            (N, 2) Array of [module ID, tpc ID] pairs, one per voxel
        meta : Meta, optional
            Meta information about the voxelized image. If provided, the
            points are assumed to be provided in voxel coordinates.

        Returns
        -------
        np.ndarray
            (N, 3) Shifted set of points
        List[np.ndarray]
            List of index of points that belong to each module
        """
        # Check that the target ID exists
        assert (
            target_id > -1 and target_id < self.tpc.num_modules
        ), "Target ID should be in [0, N_modules[."

        # Get the module ID of each of the input points
        convert = False
        if sources is not None:
            # If sources are provided, simply use that
            module_indexes = []
            for m in range(self.tpc.num_modules):
                module_indexes.append(np.where(sources[:, 0] == m)[0])

        else:
            # If the points are expressed in pixel coordinates, translate
            convert = meta is not None
            if convert:
                points = meta.to_cm(points, center=True)

            # If not provided, find module each point belongs to by proximity
            module_indexes = self.get_closest_module_indexes(points)

        # Now shifts all points that are not in the target
        for module_id, module_index in enumerate(module_indexes):
            # If this is the target module, nothing to do here
            if module_id == target_id:
                continue

            # Shift the coordinates
            points[module_index] = self.translate(
                points[module_index], module_id, target_id
            )

        # Bring the coordinates back to pixels, if they were shifted
        if convert:
            points = meta.to_px(points, floor=True)

        return points, module_indexes

    def check_containment(
        self, points, sources=None, allow_multi_module=False, summarize=True
    ):
        """Check whether a point cloud comes within some distance of the
        boundaries of a certain subset of detector volumes, depending on the
        mode.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates
        sources : np.ndarray, optional
            (S, 2) : List of [module ID, tpc ID] pairs that created the
            point cloud
        allow_multi_module : bool, default `False`
            Whether to allow points to span multiple modules
        summarize : bool, default `True`
            If `True`, only returns a single flag for the whole cloud.
            Otherwise, returns a boolean array corresponding to each point.

        Returns
        -------
        Union[bool, np.ndarray]
            `True` if the points are contained, `False` if not
        """
        # If the containment volumes are not defined, throw
        if self._cont_volumes is None:
            raise ValueError("Must call `define_containment_volumes` first.")

        # If sources are provided, only consider source volumes
        if self._cont_use_source:
            # Get the contributing TPCs
            assert len(points) == len(
                sources
            ), "Need to provide sources to make a source-based check."
            contributors = self.get_contributors(sources)
            if not allow_multi_module and len(np.unique(contributors[0])) > 1:
                return False

            # Define the smallest box containing all contributing TPCs
            # TODO: this is not ideal
            index = contributors[0] * self.tpc.num_chambers_per_module + contributors[1]
            volume = self.merge_volumes(self._cont_volumes[index])
            volumes = [volume]

        else:
            volumes = self._cont_volumes

        # Loop over volumes, make sure the cloud is contained in at least one
        if summarize:
            contained = False
            for v in volumes:
                if (points > v[:, 0]).all() and (points < v[:, 1]).all():
                    contained = True
                    break
        else:
            contained = np.zeros(len(points), dtype=bool)
            for v in volumes:
                contained |= (points > v[:, 0]).all(axis=1) & (points < v[:, 1]).all(
                    axis=1
                )

        # If requested, check points against the active volume limits
        if self._cont_limits is not None:
            if not summarize or contained:
                for norm, limit in self._cont_limits:
                    active = np.dot(points, norm) < limit
                    if summarize:
                        active = active.all()
                    contained &= active

        return contained

    def define_containment_volumes(
        self, margin, cathode_margin=None, mode="module", include_limits=True
    ):
        """This function defines a list of volumes to check containment against.

        If the containment is checked against a constant volume, it is more
        efficient to call this function once and call `check_containment`
        reapitedly after.

        Parameters
        ----------
        margin : Union[float, List[float], np.array]
            Minimum distance from a detector wall to be considered contained:
            - If float: distance buffer is shared between all 6 walls
            - If [x,y,z]: distance is shared between pairs of walls facing
              each other and perpendicular to a shared axis
            - If [[x_low,x_up], [y_low,y_up], [z_low,z_up]]: distance is
              specified individually of each wall.
        cathode_margin : float, optional
            If specified, sets a different margin for the cathode boundaries
        mode : str, default 'module'
            Containement criterion (one of 'global', 'module', 'tpc'):
            - If 'tpc', makes sure it is contained within a single TPC
            - If 'module', makes sure it is contained within a single module
            - If 'detector', makes sure it is contained within in the detector
            - If 'source', use the origin of voxels to determine which TPC(s)
              contributed to them, and define volumes accordingly
        include_limits : bool, default True
            If `True`, the TPC active region limits are checked against
        """
        # Translate the margin parameter to a (3,2) matrix
        if np.isscalar(margin):
            margin = np.full((3, 2), margin)
        elif len(np.array(margin).shape) == 1:
            assert len(margin) == 3, "Must provide one value per axis."
            margin = np.repeat([margin], 2, axis=0).T
        else:
            assert np.array(margin).shape == (3, 2), "Must provide two values per axis."
            margin = np.copy(margin)

        # Establish the volumes to check against
        self._cont_volumes = []
        if mode in ["tpc", "source"]:
            for m, module in enumerate(self.tpc):
                for t, tpc in enumerate(module):
                    vol = self.adapt_volume(
                        tpc.boundaries, margin, cathode_margin, m, t
                    )
                    self._cont_volumes.append(vol)
            self._cont_use_source = mode == "source"

        elif mode == "module":
            for module in self.tpc:
                vol = self.adapt_volume(module.boundaries, margin)
                self._cont_volumes.append(vol)
            self._cont_use_source = False

        elif mode == "detector":
            vol = self.adapt_volume(self.tpc.boundaries, margin)
            self._cont_volumes.append(vol)
            self._cont_use_source = False

        else:
            raise ValueError(f"Containement check mode not recognized: {mode}.")

        self._cont_volumes = np.array(self._cont_volumes)

        # Establish active region limits to check against, if requested
        if include_limits and self.tpc.limits is not None:
            assert np.all(margin == margin[0, 0]), (
                "No clear way to include active region limit checks when "
                "margins are different in different axes. Abort."
            )
            margin = margin[0, 0]
            self._cont_limits = []
            for limit in self.tpc.limits:
                self._cont_limits.append((limit.norm, limit.boundary - margin))

    def adapt_volume(
        self, ref_volume, margin, cathode_margin=None, module_id=None, tpc_id=None
    ):
        """Apply margins from a given volume. Takes care of subtleties
        associated with the cathode, if needed.

        Parameters
        ----------
        ref_volume : np.ndarray
            (3, 2) Array of volume boundaries
        margin : np.ndarray
            Minimum distance from a detector wall to be considered contained as
            [[x_low,x_up], [y_low,y_up], [z_low,z_up]], i.e. distance is
            specified individually of each wall.
        cathode_margin : float, optional
            If specified, sets a different margin for the cathode boundaries
        module_id : int, optional
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module

        Returns
        -------
        np.ndarray
            (3, 2) Updated array of volume boundaries
        """
        # Reduce the volume according to the margin
        volume = np.copy(ref_volume)
        volume[:, 0] += margin[:, 0]
        volume[:, 1] -= margin[:, 1]

        # If a cathode margin is provided, adapt the cathode wall differently
        if cathode_margin is not None:
            axis = self.tpc[module_id, tpc_id].drift_axis
            side = self.tpc[module_id, tpc_id].cathode_side

            flip = (-1) ** side
            volume[axis, side] += flip * (cathode_margin - margin[axis, side])

        return volume

    @staticmethod
    def merge_volumes(volumes):
        """Given a list of volumes and their boundaries, find the smallest box
        that encompass all volumes combined.

        Parameters
        ----------
        volumes : np.ndarray
            (N, 3, 2) List of volume boundaries

        Returns
        -------
        np.ndarray
            (3, 2) Boundaries of the combined volume
        """
        volume = np.empty((3, 2))
        volume[:, 0] = np.min(volumes, axis=0)[:, 0]
        volume[:, 1] = np.max(volumes, axis=0)[:, 1]

        return volume



import random
from typing import Dict, List, Union

import numpy as np
import torch
from torch import nn

from spine.data import EdgeIndexBatch, IndexBatch, TensorBatch
from spine.utils.enums import enum_factory
from spine.utils.globals import (BATCH_COL, CLUST_COL, COORD_COLS, GROUP_COL,
                                 LOWES_SHP, SHAPE_COL, TRACK_SHP)
from spine.utils.gnn.cluster import (form_clusters_batch,
                                     get_cluster_label_batch,
                                     get_cluster_primary_label_batch)
from spine.utils.gnn.evaluation import (node_assignment_batch,
                                        node_assignment_score_batch)

from .layer.common.dbscan import DBSCAN
from .layer.factories import final_factory
from .layer.gnn.factories import *

__all__ = ["GrapPA", "GrapPALoss"]


class GrapPA(torch.nn.Module):
    """Graph Particle Aggregator (GrapPA) model.

    This class mostly acts as a wrapper that will hand the graph data
    to the underlying graph neural network (GNN).

    When trained standalone, this model must be provided with a cluster
    label tensor, allowing it to build a set of intput clusters based on the
    label boundaries of the clusters and their semantic types.

    Typical configuration can look like this:

    .. code-block:: yaml

        model:
          name: grappa
          modules:
            grappa:
              nodes:
                <dictionary of arguments to specify the input type>
              graph:
                name: <name of the input graph type>
                <dictionary of arguments to specify the graph>
              node_encoder:
                name: <name of the type of node encoder>
                <dictionary of arguments to specify the node encoder>
              edge_encoder:
                name: <name of the type of edge encoder>
                <dictionary of arguments to specify the edge encoder>
              global_encoder:
                name: <name of the type of global encoder>
                <dictionary of arguments to specify the global encoder>
              gnn_model:
                name: <name of the type of backbone GNN feature extractor>
                <dictionary of arguments to specify the GNN>

    See configuration files prefixed with `grappa_` under the `config`
    directory for detailed examples of working configurations.

    See Also
    --------
    :class:`GrapPALoss`
    """

    # TODO: update
    MODULES = [
        ("grappa", ["base", "dbscan", "node_encoder", "edge_encoder", "gnn_model"]),
        "grappa_loss",
    ]

    def __init__(self, grappa, grappa_loss=None):
        """Initialize the GrapPA model.

        Parameters
        ----------
        grappa : dict
            Model configuration
        grappa_loss : dict, optional
            Loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Process the model configuration
        self.process_model_config(**grappa)

    def process_model_config(
        self,
        nodes,
        graph,
        gnn_model,
        node_encoder,
        edge_encoder=None,
        global_encoder=None,
        dbscan=None,
    ):
        """Process the top-level configuration block.

        This dispatches each block to its own configuration processor.

        Parameters
        ----------
        nodes : dict
            Input node configuration
        graph : dict
            Input graph configuration
        gnn_model : dict
            Underlying graph neural network configuration
        node_encoder : dict
            Node encoder configuration
        edge_encoder : dict
            Edge encoder configuration
        global_encoder : dict, optional
            Global encoder configuration
        dbscan : dict, optional
            DBSCAN fragmentation configuration
        """
        # Store the output types of GNNs
        self.out_types = ["node", "edge", "global"]

        # Process the node configuration
        self.process_node_config(**nodes)

        # Construct the underlying graph neural network
        self.process_gnn_config(**gnn_model)

        # Process the graph configuration
        self.graph_constructor = graph_factory(graph, self.node_type)

        # Process the encoder configurations
        self.node_encoder = node_encoder_factory(node_encoder)

        # Initialize edge encoder
        self.edge_encoder = None
        if edge_encoder is not None:
            self.edge_encoder = edge_encoder_factory(edge_encoder)

        # Initialize the global encoder
        self.global_encoder = None
        if global_encoder is not None:
            self.global_encoder = global_encoder_factory(global_encoder)

        # Process the dbscan fragmenter configuration, if provided
        self.dbscan = None
        if dbscan is not None:
            self.process_dbscan_config(dbscan)

    def process_node_config(
        self,
        source="cluster",
        shapes=None,
        min_size=-1,
        make_groups=False,
        grouping_method="score",
        grouping_through_track=False,
    ):
        """Process the node parameters of the model.

        Parameters
        ----------
        source : str, default 'cluster'
            Column name in the label tensor which contains the input cluster IDs
        shapes : int, optional
            Type of nodes to include in the input. If not specified, include
            all types
        min_size : int, default -1
            Minimum number of voxels in a cluster to be included in the input
        make_groups : bool, default False
            Use edge predictions to build node groups
        grouping_method : str, default 'score'
            Algorithm used to build a node partition
        grouping_through_track : bool, default False
            If `True`, shower objects can only be connected to one track object
        """
        # Parse the node source
        self.node_source = enum_factory("cluster", source)

        # Interpret node type as list of shapes to cluster
        if shapes is None:
            self.node_type = list(np.arange(LOWES_SHP))
        else:
            assert not np.isscalar(
                shapes
            ), "Semantic classes should be provided as a list."
            self.node_type = enum_factory("shape", shapes)

        # Store the node parameters
        self.node_min_size = min_size
        self.make_groups = make_groups
        self.grouping_method = grouping_method
        self.grouping_through_track = grouping_through_track

    def process_gnn_config(
        self, node_pred=None, edge_pred=None, global_pred=None, **gnn_model
    ):
        """Process the GNN backbone structure and the output layers.

        Parameters
        ----------
        node_pred : Union[int, dict], optional
            Number of node predictions. If there are multiple node predictions,
            provide a (key, value) pair for each type of prediction
        edge_pred : Union[int, dict], optional
            Number of edge predictions. If there are multiple edge predictions,
            provide a (key, value) pair for each type of prediction
        global_pred : Union[int, dict], optional
            Number of edge predictions. If there are multiple edge predictions,
            provide a (key, value) pair for each type of prediction
        **gnn_model, dict
            Paramters to initialize the GNN backbone
        """
        # Initialize the GNN backbone
        self.gnn = gnn_model_factory(
            gnn_model,
            node_pred is not None,
            edge_pred is not None,
            global_pred is not None,
        )

        # Initialize output layers based on the configuration
        self.process_final_config(node_pred, "node")
        self.process_final_config(edge_pred, "edge")
        self.process_final_config(global_pred, "global")

    def process_final_config(self, final, prefix):
        """Process a final layer configuration.

        Parameters
        ----------
        final : Union[int, dict]
            Final layer configuration
        prefix : dict
            Name of the final layer
        """
        # If the final layer is not specified, nothing to do here
        if final is None:
            setattr(self, f"{prefix}_pred_keys", [])
            return

        # If the final layer is specified as a number, use linear layer
        if isinstance(final, int):
            final = {"name": "linear", "out_channels": final}

        # Process the configuration dictionary otherwise
        out_keys = []
        in_channels = getattr(self.gnn, f"{prefix}_feature_size")
        if "name" in final:
            # Initialize a single final layer (single prediction of this type)
            out_key = f"{prefix}_pred"
            out_keys.append(out_key)
            setattr(self, out_key, final_factory(in_channels, **final))

        else:
            # Otherwise, initialize one final layer per prediction type
            for key, cfg in final.items():
                # If the final layer is specified as a number, use linear layer
                out_key = f"{prefix}_{key}_pred"
                out_keys.append(out_key)
                if isinstance(cfg, int):
                    cfg = {"name": "linear", "out_channels": cfg}
                setattr(self, out_key, final_factory(in_channels, **cfg))

        setattr(self, f"{prefix}_pred_keys", out_keys)

    def process_dbscan_config(shapes=None, min_size=None, **kwargs):
        """Process the DBSCAN fragmenter configuration.

        Parameters
        ----------
        shapes : Union[int, list], optional
            This should not be specified (fetched from the node configuration)
        min_size : Union[int, list], optional
            This should not be specified (fetched from the node configuration)
        **kwargs : dict, optional
            Rest of the DBSCAN configuration
        """
        # Make sure the basic parameters are not specified twice
        assert shapes is not None and min_size is not None, (
            "Do not specify 'shapes' or 'min_size' in the "
            "`dbscan` block, it is shared with the `node` block"
        )

        # Initialize DBSCAN fragmenter
        self.dbscan = DBSCAN(shapes=self.node_type, min_size=self.min_size, **kwargs)

    def forward(
        self,
        data,
        coord_label=None,
        clusts=None,
        shapes=None,
        groups=None,
        points=None,
        extra=None,
    ):
        """Prepares particle clusters and feed them to the GNN model.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is 1 (charge/energy) if the clusters (`clusts`) are provided,
              or it needs to contain cluster labels to build them on the fly
        coord_label : TensorBatch, optional
            (P, 1 + 2*D + 2) Tensor of label points (start/end/time/shape)
        clusts : IndexBatch, optional
            (C) List of indexes corresponding to each cluster
        shapes : TensorBatch, optional
            (C) List of cluster semantic class used to define the max length
        groups : TensorBatch, optional
            (C) List of node groups, one per cluster. If specified, will
                remove connections between nodes of a separate group.
        points : TensorBatch, optional
            (C, 3/6) Tensor of start (and end) points
        extra : TensorBatch, optional
            (C, N_f) Batch of features to append to the existing node features

        Returns
        -------
        clusts : IndexBatch
            (C, N_c, N_{c,i}) Cluster indexes
        edge_index : TensorBatch
            (E, 2) Incidence matrix
        node_features : TensorBatch
            (C, N_c,f) Node features
        edge_features : TensorBatch
            (C, N_e,f) Node features
        global_features : TensorBatch
            (C, N_g,f) Global features
        node_pred : TensorBatch
            (C, N_n) Node predictions (logits)
        edge_pred : TensorBatch
            (C, N_e) Edge predictions (logits)
        global_pred : TensorBatch
            (C, N_e) Global predictions (logits)
        """
        # Cast the labels to numpy for the functions run on CPU
        result = {}
        data_np = data.to_numpy()

        # If not provided, form the clusters: a list of list of voxel indices,
        # one list per cluster matching the list of requested class
        if clusts is None:
            if self.dbscan is not None:
                # Use the DBSCAN fragmenter to build the clusters
                seg_label = TensorBatch(data.tensor[:, SHAPE_COL], data.counts)
                clusts, _ = self.dbscan(data, seg_label, coord_label)
            else:
                # Use the label tensor to build the clusters
                clusts = form_clusters_batch(
                    data_np, self.node_min_size, self.node_source, shapes=self.node_type
                )

        result["clusts"] = clusts

        # If the graph edge length cut is class-specific, get the class labels
        if shapes is None and hasattr(self.graph_constructor.max_length, "__len__"):
            if self.node_source == GROUP_COL:
                # For groups, use primary shape to handle Michel/Delta properly
                shapes = get_cluster_primary_label_batch(data_np, clusts, SHAPE_COL)
            else:
                # Just use the shape of the cluster itself otherwise
                shapes = get_cluster_label_batch(data_np, clusts, SHAPE_COL)

            shapes.data = shapes.data.astype(np.int64)

        # Initialize the input graph
        edge_index, dist_mat, closest_index = self.graph_constructor(
            data_np, clusts, shapes, groups
        )

        result["edge_index"] = edge_index

        # Fetch the node features
        node_features = self.node_encoder(
            data, clusts, coord_label=coord_label, points=points, extra=extra
        )

        if isinstance(node_features, tuple):
            # If the output of the node encoder is a tuple, separate points
            node_features, points = node_features
            start_points, end_points = points.tensor.split(3, dim=1)

            result["start_points"] = TensorBatch(
                start_points, points.counts, coord_cols=np.array([0, 1, 2])
            )
            result["end_points"] = TensorBatch(
                end_points, points.counts, coord_cols=np.array([0, 1, 2])
            )

        # Fetch the edge features
        edge_features = self.edge_encoder(
            data, clusts, edge_index, closest_index=closest_index
        )

        # Feath the global_features
        global_features = None
        if self.global_encoder is not None:
            global_features = self.global_encoder(data, clusts)

        # Bring edge_index and batch_ids to device
        # TODO: try to keep everything (apart from clusts?) on GPU?
        index = torch.tensor(edge_index.index, device=data.tensor.device)
        xbatch = torch.tensor(clusts.batch_ids, device=data.tensor.device)

        # Pass through the model, update results
        out = self.gnn(node_features, index, edge_features, global_features, xbatch)

        # Loop over the necessary node/edge/global predictions, store output
        for t in self.out_types:
            for key in getattr(self, f"{t}_pred_keys"):
                result[key] = getattr(self, key)(out[f"{t}_features"])

        # If requested, build node groups from edge predictions
        if self.make_groups:
            assert (
                "edge_pred" in result
            ), "Must provide edge predictions to build node groups."
            edge_pred = result["edge_pred"].to_numpy()
            if self.grouping_method == "threshold":
                result["group_pred"] = node_assignment_batch(
                    edge_index, edge_pred, clusts
                )
            elif self.grouping_method == "score":
                if not self.grouping_through_track:
                    result["group_pred"] = node_assignment_score_batch(
                        edge_index, edge_pred, clusts
                    )
                else:
                    assert (
                        shapes is not None
                    ), "Must provide shapes to restrict track association."
                    track_node = TensorBatch(
                        shapes.data == TRACK_SHP, counts=shapes.counts
                    )
                    result["group_pred"] = node_assignment_score_batch(
                        edge_index, edge_pred, clusts, track_node
                    )

            else:
                raise ValueError(
                    "Group prediction algorithm not " "recognized:",
                    self.grouping_method,
                )

        return result


class GrapPALoss(torch.nn.modules.loss._Loss):
    """Takes the output of the GrapPA and computes the total loss.

    For use in config:

    ..  code-block:: yaml

        model:
          name: grappa
          modules:
            grappa_loss:
              node_loss:
                name: <name of the node loss>
                <dictionary of arguments to pass to the loss>
              edge_loss:
                name: <name of the edge loss>
                <dictionary of arguments to pass to the loss>
              global_loss:
                name: <name of the global loss>
                <dictionary of arguments to pass to the loss>

    Each of the specific loss blocks can also contain multiple losses by
    providing a name key in a loss block nested below it. Each loss name of a
    specific type should be provided with a corresponding output from GRaPA.

    See configuration files prefixed with `grappa_` under the `config`
    directory for detailed examples of working configurations.
    """

    def __init__(self, grappa_loss, grappa=None):
        """Initialize the GrapPA loss function.

        Parameters
        ----------
        grappa_loss : dict
            Loss configuration
        grappa : dict, optional
            Model configuration
        """
        # Initialize the parent class
        super().__init__()

        # Process the loss configuration
        self.process_loss_config(**grappa_loss)

    def process_loss_config(self, node_loss=None, edge_loss=None, global_loss=None):
        """Process the loss configuration.

        Parameters
        ----------
        node_loss : Union[dict, Dict[dict]], optional
            Node loss configuration
        edge_loss : Union[dict, Dict[dict]], optional
            Edge loss configuration
        global_loss : Union[dict, Dict[dict]], optional
            Global loss configuration
        """
        # Check that there is at least one loss to apply
        self.out_types = ["node", "edge", "global"]
        assert (
            node_loss is not None or edge_loss is not None or global_lsos is not None
        ), (
            "Must provide either a `node_loss`, `edge_loss` or "
            "`global_loss` to the GrapPA loss function."
        )

        # Initialize the node/edge/global losses
        self.process_single_loss_config("node", node_loss, node_loss_factory)
        self.process_single_loss_config("edge", edge_loss, edge_loss_factory)
        self.process_single_loss_config("global", global_loss, global_loss_factory)

    def process_single_loss_config(self, prefix, loss, constructor):
        """Process a loss configuration.

        Parameters
        ----------
        prefix : dict
            Name of the output type to apply the loss to
        loss : Union[int, dict]
            Loss configuration
        constructor : object
            Loss constructor function
        """
        # If the loss is not specified, nothing to do here
        if loss is None:
            setattr(self, f"{prefix}_loss_keys", [])
            return

        # Process the configuration dictionary otherwise
        loss_keys = []
        if "name" in loss:
            # Initialize a single loss
            loss_key = f"{prefix}_loss"
            loss_keys.append(loss_key)
            setattr(self, loss_key, constructor(loss))

        else:
            # Otherwise, initialzie one loss per prediction type
            for key, cfg in loss.items():
                loss_key = f"{prefix}_{key}_loss"
                loss_keys.append(loss_key)
                setattr(self, loss_key, constructor(cfg))

        setattr(self, f"{prefix}_loss_keys", loss_keys)

    def forward(
        self, clust_label, coord_label=None, graph_label=None, iteration=None, **output
    ):
        """Apply the node/edge/global losses to the logits from GrapPA.

        Parameters
        ----------
        clust_label : TensorBatch
            (N, 1 + D + N_f) Tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is is the number of cluster labels
        coord_label : TensorBatch, optional
            (P, 1 + D + 8) Tensor of start/end point labels for each
            true particle in the image
        graph_label : EdgeIndexTensor, optional
            (2, E) Tensor of edges that correspond to physical
            connections between true particle in the image
        iteration : int, optional
            Iteration index
        **output : dict
            Output of the GrapPA model
        """
        # Loop and apply the losses
        result = {}
        num_losses = 0
        loss, accuracy = 0.0, 0.0
        for t in self.out_types:
            loss_keys = getattr(self, f"{t}_loss_keys")
            for key in loss_keys:
                # If the number of loss keys is > 1 for this type of
                # prediction, must rename the prediction appropriately
                extra = {}
                if len(loss_keys) > 1:
                    extra[f"{t}_pred"] = output[key.replace("loss", "pred")]

                # Compute the loss
                out = getattr(self, key)(
                    clust_label=clust_label,
                    coord_label=coord_label,
                    graph_label=graph_label,
                    iteration=iteration,
                    **output,
                    **extra,
                )

                # Increment the loss and accuracy
                loss += out["loss"]
                accuracy += out["accuracy"]
                num_losses += 1

                # Update the result dictionary
                prefix = "_".join(key.split("_")[:-1])
                for k, v in out.items():
                    result[f"{prefix}_{k}"] = v

        # Append the total loss and total accuracy
        result["loss"] = torch.sum(loss) / num_losses
        result["accuracy"] = np.sum(accuracy) / num_losses

        return result



#!/usr/bin/env python3
"""Finds duplicated files."""

import argparse

import numpy as np
from larcv import larcv  # pylint: disable=W0611
from ROOT import TFile  # pylint: disable=E0611
from tqdm import tqdm


def main(source, source_list, output, tree_name):
    """Loops over a list of files and identifies files which contain the same
    set of (run, subrun, event) triplets.

    In order to save time, this script only checks if:
    1. The number of entries in the files are the same
    2. The run, subrun and event numbers in the first entry are the same

    Parameters
    ----------
    source : Union[str, List[str]]
        Path or list of paths to the input files
    source_list : str
        Path to a text file containing a list of data file paths
    output : str
        Path to the output text file with the list of duplicates
    tree_name : str
        Name of the tree to use as a reference to count the number of entries.
        If not specified, takes the first tree in the list.
    """
    # If using source list, read it in
    if source_list is not None:
        with open(source_list, "r", encoding="utf-8") as f:
            source = f.read().splitlines()

    # Initialize the output text file
    out_file = open(output, "w", encoding="utf-8")

    # Loop over the list of files in the input
    print(f"\nGathering information from {len(source)} files:")
    values = np.empty((len(source), 4), dtype=int)
    for idx, file_path in enumerate(tqdm(source)):
        # Get the tree to get the number of entries from
        f = TFile(file_path, "r")
        if tree_name is None:
            key = [key.GetName() for key in f.GetListOfKeys()][0]
        else:
            key = f"{tree_name}_tree"
        branch_key = key.replace("_tree", "_branch")

        # Check the number of entries in the file
        tree = getattr(f, key)
        num_entries = tree.GetEntries()

        # Get the event information of the first entry in the file
        tree.GetEntry(0)
        branch = getattr(tree, branch_key)
        run, subrun, event = branch.run(), branch.subrun(), branch.event()

        # Set the values list
        values[idx] = [num_entries, run, subrun, event]

    # Loop over non-unique files
    print(f"\nChecking for duplicates among {len(source)} files:")
    _, inverse, counts = np.unique(
        values, axis=0, return_inverse=True, return_counts=True
    )
    duplicate_files = []
    for idx in tqdm(np.where(counts > 1)[0]):
        # Build a file mask for this class of duplicates
        index = np.where(inverse == idx)[0]

        # All the files which are not the first in this class are duplicates
        tqdm.write(f"- File {source[index[0]]} is duplicated:")
        for i in range(1, len(index)):
            file_path = source[index[i]]
            duplicate_files.append(file_path)
            out_file.write(f"{file_path}\n")
            tqdm.write(f"  - Duplicate file: {file_path}")

    print(f"\nFound {len(duplicate_files)} duplicate files.")

    # Close text file
    out_file.close()


if __name__ == "__main__":
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description="Count entries in dataset")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--source",
        "-s",
        help="Path or list of paths to data files",
        type=str,
        nargs="+",
    )
    group.add_argument(
        "--source-list", "-S", help="Path to a text file of data file paths", type=str
    )

    parser.add_argument(
        "--output",
        "-o",
        help="Path to the output text file with the duplicate list",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--tree-name", help="TTree name used to count the entries.", type=str
    )

    args = parser.parse_args()

    # Execute the main function
    main(args.source, args.source_list, args.output, args.tree_name)



#!/usr/bin/env python3
"""Check that all the input files were processed and produced an output with
the expected number of events for each input file."""

import argparse
import os

import h5py
import numpy as np
from larcv import larcv  # pylint: disable=W0611
from ROOT import TFile  # pylint: disable=E0611
from tqdm import tqdm


def main(
    source, source_list, output, dest, suffix, event_list, tree_name, larcv_output
):
    """Checks the output of the SPINE process.

    The script loops over the input files, check that there is an output file
    in the expected location and further checks that the output file entry
    count matches that of the input file.

    Produces a list of input files that have no or incomplete output in a text
    file (the name of which is provided with the `-o` or `--output` flag). This
    can be used to reprocess missing/incomplete input files.

    .. code-block:: bash

        $ python3 bin/output_check_valid.py -S file_list.txt -o missing.txt
          --dest /path/to/output/files/ --suffix output_file_suffix

    Parameters
    ----------
    source : List[str]
        List of paths to the input files
    source_list : str
        Path to a text file containing a list of data file paths
    output : str
        Path to the output text file with the list of badly processed files
    dest : str
        Destination directory for the original SPINE process
    suffix : str
        Suffix added to the end of the input files by the original SPINE process
    event_list : str
        Path to a file containing a list of events to process. If provided, only
        events which appear on this list are required for in the output.
    tree_name : str
        Name of the tree to use as a reference to count the number of entries.
        If not specified, takes the first tree in the list.
    larcv_output, bool
        If `True`, the output file is also a ROOT file
    """
    # If using source list, read it in
    if source_list is not None:
        with open(source_list, "r", encoding="utf-8") as f:
            source = f.read().splitlines()

    # Initialize the output text file
    out_file = open(output, "w", encoding="utf-8")

    # If it is provided, parse the list of (run, subrun, event) triplets
    if event_list is not None:
        with open(event_list, "r", encoding="utf-8") as f:
            lines = f.read().splitlines()
            line_list = [l.replace(",", " ").split() for l in lines]
            event_list = [(int(r), int(s), int(e)) for r, s, e in line_list]

    # Loop over the list of files in the input
    print("\nChecking existence and completeness of output files.")
    miss_list, inc_list = [], []
    for idx, file_path in enumerate(tqdm(source)):
        # Find the base name of the input file (without extension)
        base = os.path.basename(file_path)
        stem, _ = os.path.splitext(base)

        # Check that the output exists under the expected path
        ext = ".h5" if not larcv_output else ".root"
        out_base = f"{stem}_{suffix}{ext}"
        out_path = f"{dest}/{out_base}"
        if not os.path.isfile(out_path):
            tqdm.write(f"- Missing: {out_base}")
            out_file.write(f"{file_path}\n")
            miss_list.append(file_path)
            continue

        # If the output does exist, check that the input and output have the
        # expected number of entries. If ROOT, get the tree name first.
        larcv_input = file_path.endswith(".root")
        if larcv_input:
            f = TFile(file_path, "r")
            if tree_name is None:
                key = [key.GetName() for key in f.GetListOfKeys()][0]
            else:
                key = f"{tree_name}_tree"
            key_b = key.replace("_tree", "_branch")

        # Dispatch depending if the event list is provided or not
        if event_list is None:
            # Count the number of entries in the input file
            if larcv_input:
                num_entries = getattr(f, key).GetEntries()
                f.Close()
            else:
                with h5py.File(file_path) as f:
                    num_entries = len(f["events"])

            # Then check the number of events in the output file
            if larcv_output:
                f = TFile(out_path, "r")
                out_num_entries = getattr(f, key).GetEntries()
                f.Close()
            else:
                with h5py.File(out_path) as f:
                    out_num_entries = len(f["events"])

            if out_num_entries != num_entries:
                tqdm.write(f"- Incomplete: {out_base}")
                out_file.write(f"{file_path}\n")
                inc_list.append(file_path)

        else:
            # Fetch the list of (run, subrun, event) triplets that should appear
            check_list = []
            if larcv_input:
                tree = getattr(f, key)
                for i in range(tree.GetEntries()):
                    tree.GetEntry(i)
                    branch = getattr(tree, key_b)
                    run, subrun, event = branch.run(), branch.subrun(), branch.event()
                    if (run, subrun, event) in event_list:
                        check_list.append((run, subrun, event))
                f.Close()
            else:
                with h5py.File(file_path) as f:
                    for run, subrun, event in f["run_info"]:
                        check_list.append((run, subrun, event))

            # Check that the events which should appear are present
            with h5py.File(out_path) as f:
                if len(f["events"]) != len(check_list):
                    tqdm.write(f"- Incomplete: {out_base}")
                    out_file.write(f"{file_path}\n")
                    inc_list.append(file_path)

    num_miss = len(miss_list)
    num_inc = len(inc_list)
    print(f"\nFound {num_miss + num_inc} problematic output file(s):")
    print(f"- {num_miss} missing output file(s);")
    print(f"- {num_inc} incomplete output file(s).")

    # Close text file
    out_file.close()


if __name__ == "__main__":
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description="Check dataset validity")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--source",
        "-s",
        help="Path or list of paths to data files",
        type=str,
        nargs="+",
    )
    group.add_argument(
        "--source-list", "-S", help="Path to a text file of data file paths", type=str
    )

    parser.add_argument(
        "--output",
        "-o",
        help="Path to the output text file with the problematic list",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--dest",
        help="Destination directory for the original SPINE process",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--suffix",
        help="Suffix added to the input files by the original SPINE process",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--event-list", help="File containing a list of events to process.", type=str
    )

    parser.add_argument(
        "--tree-name", help="TTree name used to count the entries.", type=str
    )

    parser.add_argument(
        "--larcv-output",
        help="The output of the process is another LArCV file",
        action="store_true",
    )

    args = parser.parse_args()

    # Execute the main function
    main(
        args.source,
        args.source_list,
        args.output,
        args.dest,
        args.suffix,
        args.event_list,
        args.tree_name,
        args.larcv_output,
    )



"""Optical detector geometry classes."""

from dataclasses import dataclass
from typing import List

import numpy as np

__all__ = ["OptDetector"]


@dataclass
class OptDetector:
    """Handles all geometry queries for a set of optical detectors.

    Attributes
    ----------
    volume : str
        The boundaries of each optical volume ('tpc' or 'module'), as defined
        by the the set of PMTs in each volume
    positions : np.ndarray
        (N_v, N_o, 3) Location of the center of each of the optical detectors
        - N_v is the number of optical volumes
        - N_o is the number of optical detectors in each volume
    shape : List[str]
        (N_d) Optical detector shape(s), combination of 'ellipsoid' and 'box'
        - N_d is the number of detector types
    dimensions : np.ndarray
        (N_d, 3) Dimensions of each of the optical detector types
        - N_d is the number of detector types
    shape_ids : np.ndarray, optional
        (N_o) Type of each of the optical detectors
        - N_o is the number of optical detectors
    det_ids : np.ndarray, optional
        (N_c) Mapping between the optical channel and its corresponding detector
        - N_c is the number of optical channels (this number can be larger
        than the number of detectors if e.g. multiple SiPMs are used per
        optical detector)
    """

    volume: str
    positions: np.ndarray
    shape: list
    dimensions: np.ndarray
    shape_ids: np.ndarray = None
    det_ids: np.ndarray = None

    def __init__(
        self,
        volume,
        volume_offsets,
        shape,
        dimensions,
        positions,
        shape_ids=None,
        det_ids=None,
        global_index=False,
        mirror=False,
    ):
        """Parse the optical detector configuration.

        Parameters
        ----------
        volume : str
            Optical decteor segmentation (per 'tpc' or per 'module')
        volume_offsets : np.ndarray
            Offsets of the optical volumes w.r.t. to the origin
        shape : Union[str, List[str]]
            Optical detector geomtry (combination of 'ellipsoid' and/or 'box')
        dimensions : Union[List[float], List[List[float]]]]
            (N_o, 3) List of optical detector dimensions along each axis
            (either the ellipsoid axis lenghts or the box edge lengths)
        positions : List[List[float]]
            (N_o, 3) Positions of each of the optical detectors
        shape_ids : List[int], optional
            (N_o) If there is different types of optical detectors, specify
            which type each of the index corresponds to
        det_ids : List[int], optional
            (N_c) If there are multiple readout channels which contribute to each
            physical optical detector, map each channel onto a physical detector
        global_index : bool, default False
            If `True`, the flash objects have a `pe_per_ch` attribute which refers
            to the entire index of optical detectors, rather than one volume
        mirror : bool, default False
            If True, mirror the z positons of the optical modules in the second
            TPC of each module
        """
        # Parse the detector shape(s) and its mapping, store is as a list
        assert shape in ["ellipsoid", "box"] or np.all(
            [s in ["ellipsoid", "box"] for s in shape]
        ), (
            "The shape of optical detectors must be represented as either "
            "an 'ellipsoid' or a 'box', or a list of them."
        )
        assert (
            isinstance(shape, str) or shape_ids is not None
        ), "If mutiple shapes are provided, must provide a shape map."
        assert shape_ids is None or len(shape_ids) == len(
            positions
        ), "Must provide a shape index for each optical channel."

        self.shape = shape
        if isinstance(shape, str):
            self.shape = [shape]
        self.shape_ids = shape_ids
        if shape_ids is not None:
            self.shape_ids = np.asarray(shape_ids, dtype=int)

        # Parse the detector dimensions, store as a 2D array
        self.dimensions = np.asarray(dimensions).reshape(-1, 3)
        assert len(self.dimensions) == len(
            self.shape
        ), "Must provide optical detector dimensions for each shape."

        # Store remaining optical detector parameters
        self.volume = volume
        self.det_ids = det_ids
        if det_ids is not None:
            self.det_ids = np.asarray(det_ids, dtype=int)

        # Parse the relative optical detector posiitons
        rel_positions = np.asarray(positions)
        if mirror:
            rel_positions_m = np.copy(rel_positions)
            rel_positions_m[:, -1] = -rel_positions_m[:, -1]

        # Store the optical detector positions in each optical volume
        count = len(positions)
        offsets = np.asarray(volume_offsets)
        self.positions = np.empty((len(offsets), count, 3))
        for v in range(len(offsets)):
            if mirror and v % 2 != 0:
                self.positions[v] = rel_positions_m + offsets[v]
            else:
                self.positions[v] = rel_positions + offsets[v]

        # Store if the flash points to the entire index of optical detectors
        self.global_index = global_index

    @property
    def num_volumes(self):
        """Returns the number of optical volumes.

        Returns
        -------
        int
            Number of optical volumes, N_v
        """
        return self.positions.shape[0]

    @property
    def num_detectors_per_volume(self):
        """Returns the number of optical detectors in each optical volume.

        Returns
        -------
        int
            Number of optical detectors in each volume, N_o
        """
        return self.positions.shape[1]

    @property
    def num_detectors(self):
        """Number of optical detectors.

        Returns
        -------
        int
            Total number of optical detector, N_v*N_o
        """
        return self.num_volumes * self.num_detectors_per_volume

    def volume_index(self, volume_id):
        """Returns an index which corresponds to detectors in a certain volume.

        Parameters
        ----------
        volume_id : int
            ID of the volume to return the index for

        Returns
        -------
        np.ndarray
            Index of the detectors which belong to the requested volume ID
        """
        # If using a global index, all volumes point to the same index
        if self.global_index:
            return np.arange(self.num_detectors)

        return volume_id * self.num_detectors_per_volume + np.arange(
            self.num_detectors_per_volume
        )



"""Analysis script used to evaluate the point proposal accuracy."""

from warnings import warn

import numpy as np
from scipy.spatial.distance import cdist
from scipy.special import softmax

from spine.ana.base import AnaBase
from spine.utils.globals import (COORD_COLS, LOWES_SHP, PPN_END_COLS,
                                 PPN_LENDP_COL, PPN_LTYPE_COL, PPN_SHAPE_COL)

__all__ = ["PointProposalAna"]


class PointProposalAna(AnaBase):
    """Class which computes and stores the necessary data to evaluate the
    point proposal accuracy.

    It evaluates the following metrics:
    - Distance from true to closest predicted point (efficiency)
    - Distance from predicted to closest true point (purity)
    - Point type classification accuracy
    - Point end classification accuracy
    """

    # Name of the analysis script (as specified in the configuration)
    name = "point_eval"

    # Set of data keys needed for this analysis script to operate
    _keys = (("ppn_pred", True),)

    def __init__(
        self, num_classes=LOWES_SHP, label_key="ppn_label", endpoints=False, **kwargs
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        num_classes : int, default 4
            Number of pixel classses, excluding the ghost class
        label_key : str, default 'seg_label'
            Name of the tensor which contains the segmentation labels
        endpoints : bool, default False
            Evaluate the accuracy of end point classification
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the basic parameters
        self.num_classes = num_classes
        self.label_key = label_key
        self.endpoints = endpoints

        # Append other required key
        self.update_keys({self.label_key: True})

        # Initialize the output
        self.initialize_writer("truth_to_reco")
        self.initialize_writer("reco_to_truth")

        # Initialize a dummy dictionary to return when there is no match
        self.dummy_dict = {"dist": -1.0, "shape": -1, "closest_shape": -1}
        for s in range(self.num_classes):
            self.dummy_dict[f"dist_{s}"] = -1.0
        if endpoints:
            self.dummy_dict["end"] = -1
            self.dummy_dict["closest_end"] = -1

    def process(self, data):
        """Store the semantic segmentation metrics for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Initialize dictionaries to compare
        points, types, ends = {}, {}, {}

        # Fetch the list of label points and their characteristics
        points["truth"] = data[self.label_key][:, COORD_COLS]
        types["truth"] = data[self.label_key][:, PPN_LTYPE_COL].astype(int)
        if self.endpoints:
            ends["truth"] = data[self.label_key][:, PPN_LENDP_COL].astype(int)

        # Fetch the list of predicted points and their characteristics
        points["reco"] = data["ppn_pred"][:, COORD_COLS]
        types["reco"] = data["ppn_pred"][:, PPN_SHAPE_COL].astype(int)
        if self.endpoints:
            ends["reco"] = np.argmax(data["ppn_pred"][:, PPN_END_COLS], axis=1)

        # Compute the pair-wise distances between label and predicted points
        dist_mat = {}
        dist_mat["reco"] = cdist(points["reco"], points["truth"])
        dist_mat["truth"] = dist_mat["reco"].T

        # Store one row per predicted point
        matches = (("truth", "reco"), ("reco", "truth"))
        for k, (source, target) in enumerate(matches):
            # If there is no source points, nothing to do
            if not len(points[source]):
                continue

            # If there are no target points, record no match
            if not len(points[target]):
                # Append dummy values
                dummy = {**self.dummy_dict}
                for i in range(len(points[source])):
                    dummy["shape"] = types[source][i]
                    if self.endpoints:
                        dummy["end"] = ends[source][i]
                    self.append(f"{source}_to_{target}", **dummy)

                # Proceed
                continue

            # Otherwise, use closest point as reference
            dists = dist_mat[source]
            closest_index = np.argmin(dists, axis=1)
            masks = [np.where(types[target] == s)[0] for s in range(self.num_classes)]
            for i in range(len(points[source])):
                point_dict = {**self.dummy_dict}
                point_dict["dist"] = dists[i, closest_index[i]]
                point_dict["shape"] = types[source][i]
                point_dict["closest_shape"] = types[target][closest_index[i]]
                if self.endpoints:
                    point_dict["end"] = ends[source][i]
                    point_dict["closest_end"] = ends[target][closest_index[i]]
                for s in range(self.num_classes):
                    if len(masks[s]) > 0:
                        point_dict[f"dist_{s}"] = np.min(dists[i, masks[s]])

                self.append(f"{source}_to_{target}", **point_dict)



#!/usr/bin/env python3
"""Counts the number of events in a LArCV dataset."""

import argparse

from larcv import larcv  # pylint: disable=W0611
from ROOT import TFile  # pylint: disable=E0611
from tqdm import tqdm


def main(source, source_list, tree_name):
    """Checks the number of entries in a file/list of files.

    Parameters
    ----------
    source : Union[str, List[str]]
        Path or list of paths to the input files
    source_list : str
        Path to a text file containing a list of data file paths
    tree_name : str
        Name of the tree to use as a reference to count the number of entries.
        If not specified, takes the first tree in the list.
    """
    # If using source list, read it in
    if source_list is not None:
        with open(source_list, "r", encoding="utf-8") as f:
            source = f.read().splitlines()

    # Loop over the list of files in the input
    total_entries = 0
    print(f"\nCounting entries in {len(source)} file(s):")
    for file_path in tqdm(source):
        # Get the tree to get the number of entries from
        f = TFile(file_path, "r")
        if tree_name is None:
            key = [key.GetName() for key in f.GetListOfKeys()][0]
        else:
            key = f"{tree_name}_tree"

        # Count the number of entries in this file
        num_entries = getattr(f, key).GetEntries()
        f.Close()

        # Dump number for this file, increment
        tqdm.write(f"- Counted {num_entries} entries in {file_path}")
        total_entries += num_entries

    print(f"\nTotal number of entries: {total_entries}")


if __name__ == "__main__":
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description="Count entries in dataset")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--source",
        "-s",
        help="Path or list of paths to data files",
        type=str,
        nargs="+",
    )
    group.add_argument(
        "--source-list", "-S", help="Path to a text file of data file paths", type=str
    )

    parser.add_argument(
        "--tree-name", help="TTree name used to count the entries.", type=str
    )

    args = parser.parse_args()

    # Execute the main function
    main(args.source, args.source_list, args.tree_name)



"""Basic detector components shared across multiple subsystems.

This currently handles:
- :class:`Box` which corresponds to box-shaped detector modules.
"""

from dataclasses import dataclass

import numpy as np

__all__ = ["Box", "Plane"]


@dataclass
class Box:
    """Class which holds all methods associated with a box-shaped component.

    Attributes
    ----------
    boundaries : np.ndarray
        (3, 2) Box boundaries
        - 3 is the number of dimensions
        - 2 corresponds to the lower/upper boundaries along each axis
    """

    boundaries: np.ndarray

    def __init__(self, lower, upper):
        """Initialize the box object.

        Parameters
        ----------
        lower : np.ndarray
            (3) Lower bounds of the box
        upper : np.ndarray
            (3) Upper bounds of the box
        """
        # Store lower and upper boundaries in one array
        self.boundaries = np.vstack((lower, upper)).T

    @property
    def center(self):
        """Center of the box.

        Returns
        -------
        np.ndarray
            Center of the box
        """
        return np.mean(self.boundaries, axis=1)

    @property
    def lower(self):
        """Lower bounds of the box.

        Returns
        -------
        np.ndarray
            Lower bounds of the box
        """
        return self.boundaries[:, 0]

    @property
    def upper(self):
        """Upper bounds of the box.

        Returns
        -------
        np.ndarray
            Upper bounds of the box
        """
        return self.boundaries[:, 1]

    @property
    def dimensions(self):
        """Dimensions of the box.

        Returns
        -------
        np.ndarray
            Box dimensions
        """
        return self.boundaries[:, 1] - self.boundaries[:, 0]


@dataclass
class Plane:
    """Class which holds all methods associated with a plane.

    Attributes
    ----------
    intercept : np.ndarray
        (3) Coordinates of a point which belongs to the plane
    norm : np.ndarray
        (3) Vector perpendicular to the plane
    boundary : float
        Dot product between the interecept and the norm
    """

    intercept: np.ndarray
    norm: np.ndarray
    boundary: float

    def __init__(self, intercept, norm):
        """Initialize the box object.

        Parameters
        ----------
        intercept : np.ndarray
            (3) Coordinates of a point which belongs to the plane
        norm : np.ndarray
            (3) Vector perpendicular to the plane
        """
        # Store the intercept and norm
        self.intercept = np.asarray(intercept)
        self.norm = np.asarray(norm)

        # Compute their dot product (projection boundary)
        self.boundary = np.dot(intercept, norm)



from collections import OrderedDict

from spine.ana.base import AnaBase

__all__ = []


# @write_to(['pid_metrics'])
def pid_metrics(
    data_dict,
    result_dict,
    iteration=-1,
    logger=None,
    matching_mode="pred_to_true",
    primary_only=True,
    mpv_only=True,
):
    """
    Script which stores the scores, predictions and labels related
    to particle identification in order to evaluate PID performance.

    Parameters
    ----------
    data_dict : dict
        Input data dictionary after post-processing
    result_dict : dict
        Chain output dictionary after post-processing
    primary_only : bool, default True
        If True, only store particles corresponding to primary particles
    mpv_only : bool, default True
        If True, only store particles corresponding to particles from MPV interactions

    Returns
    -------
    List[dict]
        One dictionary of relevant particle attribute per particle in
        the image being processed.

    Information in <pid_metrics> will be saved to $log_dir/pid_metrics.csv.
    """
    # Initialize index dictionary to be shared by all entries
    index_dict = {"iteration": iteration, "index": data_dict["index"][0]}

    # Get list of matched particle objects
    assert matching_mode in [
        "true_to_pred",
        "pred_to_true",
    ], f"Matching mode is not recognized: {matching_mode}"

    if "matched_particles" in result_dict:
        matches, counts = (
            result_dict["matched_particles"][0],
            result_dict["particle_match_overlap"][0],
        )
    else:
        suffix = "t2r" if matching_mode == "true_to_pred" else "r2t"
        matches, counts = (
            result_dict[f"matched_particles_{suffix}"][0],
            result_dict[f"particle_match_overlap_{suffix}"][0],
        )

    # Loop over matches
    metrics = []
    for i, mparticles in enumerate(matches):
        # If there is no principal match, skip
        if mparticles[1] is None:
            continue

        # Fetch particle objects
        if matching_mode == "true_to_pred":
            true_p, pred_p = mparticles[0], mparticles[1]
            n_classes = len(true_p.pid_scores)
        elif matching_mode == "pred_to_true":
            pred_p, true_p = mparticles[0], mparticles[1]
            n_classes = len(pred_p.pid_scores)

        assert (type(true_p) is TruthParticle) or (true_p) is None
        assert (type(pred_p) is Particle) or (pred_p) is None

        # Apply restrictions, if requested
        if primary_only and not true_p.is_primary:
            continue
        if mpv_only and not true_p.nu_id > -1:
            continue

        # Build dictionary
        metrics_dict = OrderedDict()
        metrics_dict.update(index_dict)
        metrics_dict["true_id"] = true_p.id
        metrics_dict["pred_id"] = pred_p.id
        metrics_dict["is_primary"] = int(true_p.is_primary)
        metrics_dict["nu_id"] = true_p.nu_id
        metrics_dict["match_overlap"] = counts[i]
        metrics_dict["label"] = true_p.pid
        metrics_dict["pred"] = pred_p.pid
        for c in range(n_classes):
            metrics_dict[f"score_{c}"] = pred_p.pid_scores[c]

        # Append
        metrics.append(metrics_dict)

    return [metrics]



from collections import defaultdict
from itertools import combinations

import numpy as np
from analysis.post_processing.reconstruction.utils import \
    closest_distance_two_lines

from spine.utils.gnn.cluster import cluster_direction

# TODO: Need to refactor according to post processing conventions


def _tag_neutral_pions_true(particles):
    out = []
    tagged = defaultdict(list)
    for part in particles:
        num_voxels_noghost = part.coords_noghost.shape[0]
        ancestor = part.ancestor_track_id
        if (
            part.pdg_code == 22
            and part.creation_process == "Decay"
            and part.parent_creation_process == "primary"
            and part.ancestor_pdg_code == 111
            and num_voxels_noghost > 0
        ):
            tagged[ancestor].append(part.id)
    for photon_list in tagged.values():
        out.append(tuple(photon_list))
    return out


def _tag_neutral_pions_reco(particles, threshold=5):
    out = []
    photons = [p for p in particles if p.pid == 0]
    for entry in combinations(photons, 2):
        p1, p2 = entry
        v1, v2 = cluster_direction(p1), cluster_direction(p2)
        d = closest_distance_two_lines(p1.startpoint, v1, p2.startpoint, v2)
        if d < threshold:
            out.append((p1.id, p2.id))
    return out


def tag_neutral_pions(particles, mode):
    if mode == "truth":
        return _tag_neutral_pions_true(particles)
    elif mode == "pred":
        return _tag_neutral_pions_reco(particles)
    else:
        raise ValueError


def closest_distance_two_lines(a0, u0, a1, u1):
    """
    a0, u0: point (a0) and unit vector (u0) defining line 1
    a1, u1: point (a1) and unit vector (u1) defining line 2
    """
    cross = np.cross(u0, u1)
    # if the cross product is zero, the lines are parallel
    if np.linalg.norm(cross) == 0:
        # use any point on line A and project it onto line B
        t = np.dot(a1 - a0, u1)
        a = a1 + t * u1  # projected point

        return np.linalg.norm(a0 - a)
    else:
        # use the formula from https://en.wikipedia.org/wiki/Skew_lines#Distance
        t = np.dot(np.cross(a1 - a0, u1), cross) / np.linalg.norm(cross) ** 2

        # closest point on line A to line B
        p = a0 + t * u0

        # closest point on line B to line A
        q = p - cross * np.dot(p - a1, cross) / np.linalg.norm(cross) ** 2

        return np.linalg.norm(p - q)  # distance between p and q



#!/usr/bin/env python3
"""Builds a list of file which make a data run."""

import argparse

from larcv import larcv  # pylint: disable=W0611
from ROOT import TFile  # pylint: disable=E0611
from tqdm import tqdm


def main(source, source_list, output, run_number, tree_name):
    """Loops over a list of files and finds those which belong to a certain run.

    Parameters
    ----------
    source : Union[str, List[str]]
        Path or list of paths to the input files
    source_list : str
        Path to a text file containing a list of data file paths
    output : str
        Path to the output text file with the list of run files
    run_number : int
        Run number to look for
    tree_name : str
        Name of the tree to use as a reference to get the run number from.
        If not specified, takes the first tree in the list.
    """
    # If using source list, read it in
    if source_list is not None:
        with open(source_list, "r", encoding="utf-8") as f:
            source = f.read().splitlines()

    # Initialize the output text file
    out_file = open(output, "w", encoding="utf-8")

    # Loop over the list of files in the input
    print(f"\nLooking for run {run_number} in {len(source)} files:")
    run_files = []
    for file_path in tqdm(source):
        # Get the tree to get the number of entries from
        f = TFile(file_path, "r")
        if tree_name is None:
            key = [key.GetName() for key in f.GetListOfKeys()][0]
        else:
            key = f"{tree_name}_tree"
        branch_key = key.replace("_tree", "_branch")

        # Check the run number of the first entry in the file
        tree = getattr(f, key)
        tree.GetEntry(0)
        run = getattr(tree, branch_key).run()
        f.Close()

        # If the file contains entries from the correct run, append
        if run == run_number:
            tqdm.write(f"- Good file: {file_path}")
            run_files.append(file_path)
            out_file.write(f"{file_path}\n")

    print(f"\nFound {len(run_files)} run {run_number} files.")

    # Close text file
    out_file.close()


if __name__ == "__main__":
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description="Count entries in dataset")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--source",
        "-s",
        help="Path or list of paths to data files",
        type=str,
        nargs="+",
    )
    group.add_argument(
        "--source-list", "-S", help="Path to a text file of data file paths", type=str
    )

    parser.add_argument(
        "--output",
        "-o",
        help="Path to the output text file with the run file list",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--run-number", help="Run number to look for", type=int, required=True
    )

    parser.add_argument(
        "--tree-name", help="TTree name used to count the entries.", type=str
    )

    args = parser.parse_args()

    # Execute the main function
    main(args.source, args.source_list, args.output, args.run_number, args.tree_name)



"""Script which injects a run number in every event of every tree in a file or
a list of files.
"""

import argparse
import os
import tempfile

import numpy as np
from larcv import larcv  # pylint: disable=W0611
from ROOT import TFile  # pylint: disable=E0611
from tqdm import tqdm

# LArCV IO Manager configuration string
CFG = """
IOManager: {
    Verbosity   : 4
    Name        : "OutIO"
    IOMode      : 2
    InputFiles  : [INPUT_PATH]
    OutFileName : OUTPUT_PATH
}
"""


def initialize_manager(file_path, dest, overwrite, suffix):
    """Initialize an IOManager object given a configuration.

    Parameters
    ----------
    file_path : str
        Path to the input file

    Returns
    -------
    larcv.IOManager
        IOManager object
    """
    # If the destination is provided, direct the output file there
    out_path = file_path
    if dest is not None:
        base = os.path.basename(file_path)
        out_path = f"{dest}/{base}"

    # If a suffix is provided, append
    assert (
        suffix is None or not overwrite
    ), "No point in providing a suffix if the original file is overwritten."
    if suffix is not None:
        out_path = out_path.replace(".root", f"_{suffix}.root")
    elif overwrite:
        out_path = out_path.replace(".root", "_tmp.root")

    # Check that the output file does is not the same as the original file
    if file_path == out_path:
        raise ValueError(
            "The input file name and the output file name are the same. "
            "This is not allowed by the LArCV IOManager."
        )

    # Update the configuration with the input/output file names
    cfg = CFG
    cfg = cfg.replace("INPUT_PATH", file_path)
    cfg = cfg.replace("OUTPUT_PATH", out_path)

    # Create a temporary text file with the configuration
    tmp = tempfile.NamedTemporaryFile("w")
    tmp.write(cfg)
    tmp.flush()

    # Initialize the IOManager
    manager = larcv.IOManager(tmp.name)
    manager.initialize()

    return manager, out_path


def main(source, source_list, dest, overwrite, run_number, suffix):
    """Checks the output of the SPINE process.

    The script loops over the input files, fetch the list of keys in the file
    and injects a run number of each event in each file.

    .. code-block:: bash

        $ python3 bin/inject_run_number.py -S file_list.txt
          --overwrite --run_number 123

    Parameters
    ----------
    source : List[str]
        List of paths to the input files
    source_list : str
        Path to a text file containing a list of data file paths
    dest : str
        Destination folder to write the files to
    overwrite : bool
        If `True`, overwrite the original files
    run_number : int
        Run number to inject in the input file list. If it is specied as -1,
        each file is assigned a unique run number
    suffix : str
        String to append to the end of the input file names to form the name
        of the output file with the updated run numbers
    """
    # If using source list, read it in
    if source_list is not None:
        with open(source_list, "r", encoding="utf-8") as f:
            source = f.read().splitlines()

    # Loop over the list of files in the input
    print("\nUpdating the run numbers of input files.")
    for idx, file_path in enumerate(tqdm(source)):
        # Initialize the input/output processes
        io, out_path = initialize_manager(file_path, dest, overwrite, suffix)

        # Loop over entries, set the run number for every data product
        num_entries = io.get_n_entries()
        run = run_number if run_number > -1 else idx
        for e in range(num_entries):
            # Read existing content
            io.read_entry(e)

            # Update the run number
            io.set_id(run, 0, e + 1)

            # Save
            io.save_entry()

        # Finalize
        io.finalize()

        # If needed move the output file to where the
        if overwrite:
            os.rename(out_path, file_path)


if __name__ == "__main__":
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description="Check dataset validity")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--source",
        "-s",
        help="Path or list of paths to data files",
        type=str,
        nargs="+",
    )
    group.add_argument(
        "--source-list", "-S", help="Path to a text file of data file paths", type=str
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--dest", help="Destination folder for the output file", type=str
    )
    group.add_argument(
        "--overwrite",
        help="Overwrite the input file with the output file",
        action="store_true",
    )

    parser.add_argument(
        "--run-number",
        help="Run number to assign to every input file",
        type=int,
        required=True,
    )

    parser.add_argument(
        "--suffix", help="Suffix to append to the input file names", type=str
    )

    args = parser.parse_args()

    # Execute the main function
    main(
        args.source,
        args.source_list,
        args.dest,
        args.overwrite,
        args.run_number,
        args.suffix,
    )



"""Analysis script used to store basic event information into CSV files."""

import copy
from collections import defaultdict

import numpy as np

from spine.ana.base import AnaBase
from spine.utils.globals import ELEC_PID, SHOWR_SHP

__all__ = ["EventAna"]


class EventAna(AnaBase):
    """Class which saves basic event information (and their matches)."""

    # Name of the analysis script (as specified in the configuration)
    name = "event"

    # Set of data keys needed for this analysis script to operate
    _keys = (("reco_particles", True), ("truth_particles", True))

    def __init__(self, **kwargs):
        """Initialize the CSV event logging class.

        Parameters
        ----------
        **kwargs : dict, optional
            Parameters to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Initialize the output log file
        self.initialize_writer("events")

    def process(self, data):
        """Store basic event information for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products containing particle representations
        """
        # Get the basic event-level information
        # TODO: this should be using run_mode
        reco_p, truth_p = data["reco_particles"], data["truth_particles"]
        cnts = defaultdict(int)
        for out_type in ["reco", "truth"]:
            particles = data[f"{out_type}_particles"]
            pre = f"num_{out_type}"
            for p in particles:
                cnts[f"{pre}_particles"] += 1
                cnts[f"{pre}_primaries"] += p.is_primary
                cnts[f"{pre}_voxels"] += p.size
                cnts[f"{pre}_showers"] += p.shape == SHOWR_SHP
                cnts[f"{pre}_primary_showers"] += p.shape == SHOWR_SHP and p.is_primary
                cnts[f"{pre}_electron_showers"] += p.pid == ELEC_PID

        row_dict.update(**self.base_dict, **cnts)
        self.writers["events"].append(row_dict)



"""Module that contains parsers that do not fit in other categories.

Contains the following parsers:
- :class:`Meta2DParser`
- :class:`Meta3DParser`
- :class:`RunInfoParser`
- :class:`CRTHitParser`
- :class:`TriggerParser`
"""

from spine.data import CRTHit, Flash, Meta, RunInfo, Trigger
from spine.utils.conditional import larcv
from spine.utils.optical import FlashMerger

from .base import ParserBase
from .data import ParserObjectList

__all__ = [
    "MetaParser",
    "RunInfoParser",
    "FlashParser",
    "CRTHitParser",
    "TriggerParser",
]


class MetaParser(ParserBase):
    """Get the metadata information to translate into real world coordinates.

    Each entry in a dataset is a cube, where pixel/voxel coordinates typically
    go from 0 to some integer N in each dimension. If you wish to translate
    these pixel/voxel coordinates back into real world coordinates, you can use
    the output of this parser to compute it.

    .. code-block. yaml

        schema:
          meta:
            parser: meta
            sparse_event: sparse3d_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "meta"

    # Alternative allowed names of the parser
    aliases = ("meta2d", "meta3d")

    # Type of object(s) returned by the parser
    returns = "object"

    # Overlay strategy for the objects returned by the parser
    overlay = "match"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def __init__(self, projection_id=None, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        projection_id : int, optional
            Projection ID to get the 2D image from (if fetching from 2D)
        **kwargs : dict, optional
            data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the revelant attributes
        self.projection_id = projection_id

    def process(self, sparse_event=None, cluster_event=None):
        """Fetches the metadata from one object that has it.

        Parameters
        ----------
        sparse_event : Union[larcv.EventSparseTensor2D
                             larcv.EventSparseTensor3D], optional
            Tensor which contains the metadata information as an attribute
        cluster_event : Union[larcv.EventClusterPixel2D,
                              larcv.EventClusterVoxel3D], optional
            Cluster which contains the metadata information as an attribute

        Returns
        -------
        Meta
            Metadata information for one image
        """
        # Check on the input, pick a source for the metadata
        assert (sparse_event is not None) ^ (
            cluster_event is not None
        ), "Must specify either `sparse_event` or `cluster_event`."
        ref_event = sparse_event if sparse_event is not None else cluster_event

        # Fetch a specific projection, if needed
        if isinstance(
            ref_event, (larcv.EventSparseTensor2D, larcv.EventClusterPixel2D)
        ):
            ref_event = ref_event.sparse_tensor_2d(self.projection_id)

        return Meta.from_larcv(ref_event.meta())


class RunInfoParser(ParserBase):
    """Parse run information (run, subrun, event number).

    .. code-block. yaml

        schema:
          run_info:
            parser: run_info
            sparse_event: sparse3d_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "run_info"

    # Type of object(s) returned by the parser
    returns = "object"

    # Overlay strategy for the objects returned by the parser
    overlay = "cat"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, sparse_event=None, cluster_event=None):
        """Fetches the run information from one object that has it.

        Parameters
        ----------
        sparse_event : Union[larcv.EventSparseTensor2D
                             larcv.EventSparseTensor3D], optional
            Tensor which contains the run information as an attribute
        cluster_event : Union[larcv.EventClusterPixel2D,
                              larcv.EventClusterVoxel3D], optional
            Cluster which contains the run information as an attribute

        Returns
        -------
        RunInfo
            Run information object
        """
        # Check on the input, pick a source for the run information
        assert (sparse_event is not None) ^ (
            cluster_event is not None
        ), "Must specify either `sparse_event` or `cluster_event`."
        ref_event = sparse_event if sparse_event is not None else cluster_event

        return RunInfo.from_larcv(ref_event)


class FlashParser(ParserBase):
    """Copy construct Flash and return an array of `Flash`.

    This parser also takes care of flashes that have been split between their
    respective optical volumes, provided a `flash_event_list`. This parser
    assumes that the trees are provided in order of the volume ID they
    correspond to.

    .. code-block. yaml
        schema:
          flashes:
            parser: flash
            flash_event_list:
              - flash_cryoE
              - flash_cryoW
    """

    # Name of the parser (as specified in the configuration)
    name = "flash"

    # Alternative allowed names of the parser
    aliases = ("opflash",)

    # Type of object(s) returned by the parser
    returns = "object_list"

    def __init__(self, merge=None, **kwargs):
        """Initialize the flash parser.

        Parameters
        ----------
        merge : dict, optional
            Flash merging configuration
        **kwargs : dict, optional
            data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Initialize the flash merging class, if needed
        self.merger = None
        if merge is not None:
            self.merger = FlashMerger(**merge)

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, flash_event=None, flash_event_list=None):
        """Fetches the list of optical flashes.

        Parameters
        -------------
        flash_event : larcv.EventFlash, optional
            Optical flash event which contains a list of flash objects
        flash_event_list : List[larcv.EventFlash], optional
            List of optical flash events, each a list of flash objects

        Returns
        -------
        List[Flash]
            List of optical flash objects
        """
        # Check on the input
        assert (flash_event is not None) ^ (
            flash_event_list is not None
        ), "Must specify either `flash_event` or `flash_event_list`."

        # Parse flash objects
        if flash_event is not None:
            # If there is a single flash event, parse it as is
            flash_list = flash_event.as_vector()
            flashes = [Flash.from_larcv(larcv.Flash(f)) for f in flash_list]

        else:
            # Otherwise, set the volume ID of the flash to the source index
            # and count the flash index from 0 to the largest number
            flashes = []
            idx = 0
            for volume_id, flash_event in enumerate(flash_event_list):
                for f in flash_event.as_vector():
                    # Cast and update attributes
                    flash = Flash.from_larcv(f)
                    flash.id = idx
                    flash.volume_id = volume_id

                    # Append, increment counter
                    flashes.append(flash)
                    idx += 1

        # If requested, merge flashes which match in time
        if self.merger is not None:
            flashes, _ = merger(flashes)

        return ParserObjectList(flashes, Flash())


class CRTHitParser(ParserBase):
    """Copy construct CRTHit and return an array of `CRTHit`.

    .. code-block. yaml
        schema:
          crthits:
            parser: crthit
            crthit_event: crthit_crthit
    """

    # Name of the parser (as specified in the configuration)
    name = "crthit"

    # Type of object(s) returned by the parser
    returns = "object_list"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, crthit_event):
        """Fetches the list of CRT hits.

        Parameters
        ----------
        crthit_event : larcv.CRTHitEvent

        Returns
        -------
        List[CRTHit]
            List of CRT hit objects
        """
        # Output as a list of LArCV CRT hit objects
        crthit_list = crthit_event.as_vector()
        crthits = [CRTHit.from_larcv(larcv.CRTHit(c)) for c in crthit_list]

        return ParserObjectList(crthits, CRTHit())


class TriggerParser(ParserBase):
    """Copy construct Trigger and return a `Trigger`.

    .. code-block. yaml
        schema:
          trigger:
            parser: trigger
            trigger_event: trigger_base
    """

    # Name of the parser (as specified in the configuration)
    name = "trigger"

    # Type of object(s) returned by the parser
    returns = "object"

    # Overlay strategy for the objects returned by the parser
    overlay = "cat"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, trigger_event):
        """Fetches the trigger information.

        Parameters
        ----------
        trigger_event : larcv.TriggerEvent

        Returns
        -------
        Trigger
            Trigger object
        """
        # Output as a trigger objects
        trigger = Trigger.from_larcv(larcv.Trigger(trigger_event))

        return trigger
